import numpy as np

"""
modules/xrenner_classify.py

Adapter class to accommodate different types of classifiers. 
  * Supplies a uniform predict_proba() method accessed via .classify_many()
  * Reads pickled Encoders/Binarizers/Scalers and ensures encoding of OOV feature values as _unknown_
  * Compatible with several prominent sklearn estimators (e.g. GradientBoosting, RandomForest, MLP and linear models)

Author: Amir Zeldes
"""

class Classifier:
	def __init__(self, cls, encoder_dict, headers):
		self.cls = cls
		self.encoder_dict = encoder_dict
		self.headers = headers
		t = str(type(cls))
		if "Ridge" in t or "Elastic" in t or "Logistic" in t:
			self.cls_type = "decision"
		elif "RandomForest" in t or "Perceptron" in t or "StochasticGradient" in t or "Boost" in t or \
			"XGB" in t:
			self.cls_type = "tuple"
		else:
			self.cls_type = "predict_proba"


	def classify_many(self, input_list):
		"""
		Classifies a list of xrenner anaphor-antecedent candidate dump features

		:param input_list: list of numpy arrays containing markable pair features
		:return: numpy array of coref probabilities for the input pairs
		"""

		encoded_arrays = []
		for ana, ante, candidate_list, lex in input_list:
			feats = ana.extract_features(lex, ante, candidate_list)
			encoded = []
			for header in self.headers:
				if header in self.encoder_dict:  # Categorical feature
					if self.encoder_dict[header][1] == "binarizer":
						if feats[header] in self.encoder_dict[header][2]:
							cols = self.encoder_dict[header][0].transform([feats[header]])
							encoded += cols.tolist()[0]
						else:  # OOV item
							cols = self.encoder_dict[header][0].transform(["_unknown_"])
							encoded += cols.tolist()[0]
					elif self.encoder_dict[header][1] == "scale":
						encoded.append(self.encoder_dict[header][0].transform(np.array(feats[header])))
					else:  # ordinal feature
						if feats[header] in self.encoder_dict[header][2]:
							encoded.append(self.encoder_dict[header][0].transform(np.array([feats[header]]).reshape(-1,1)))
						else:  # OOV item
							encoded.append(self.encoder_dict[header][0].transform(np.array(["_unknown_"]).reshape(-1,1)))
				else:  # Untransformed numerical feature
					encoded.append(feats[header])
			encoded_array = np.array(encoded).reshape((1, -1))
			encoded_arrays.append(encoded_array)
		return self.predict_proba(np.array(encoded_arrays).reshape(len(input_list),-1))

	def predict_proba(self, feat_matrix):
		if self.cls_type == "predict_proba":
			return self.cls.predict_proba(feat_matrix)
		elif self.cls_type == "decision":  # Returns proba via decision function, get exp
			d = self.cls.decision_function(feat_matrix)#[0]
			probs = np.exp(d) / (1 + np.exp(d))
			return probs
		elif self.cls_type == "tuple":  # Returns tuple with class probabilities for each case
			probas = self.cls.predict_proba(feat_matrix)
			return [tpl[1] for tpl in probas]







import sys, io, os, re, tempfile, subprocess
from depedit import DepEdit

script_dir = os.path.dirname(os.path.realpath(__file__)) + os.sep
bin_dir = script_dir + ".." + os.sep + "bin" + os.sep
data_dir = script_dir + ".." + os.sep + "data" + os.sep
parser_path = bin_dir + "maltparser-1.8" + os.sep

PY3 = sys.version_info[0] == 3


def exec_via_temp(input_text, command_params, workdir=""):
	temp = tempfile.NamedTemporaryFile(delete=False)
	exec_out = ""
	try:
		temp.write(input_text.encode("utf8"))
		temp.close()

		command_params = [x if x != 'tempfilename' else temp.name for x in command_params]
		if workdir == "":
			proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE)
			(stdout, stderr) = proc.communicate()
		else:
			proc = subprocess.Popen(command_params, stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
			(stdout, stderr) = proc.communicate()

		exec_out = stdout
	except Exception as e:
		print(e)
	finally:
		os.remove(temp.name)
		if PY3:
			exec_out = exec_out.decode("utf8")
		return exec_out


d = DepEdit(config_file=data_dir + "add_ud_and_flat_morph.ini",options=type('', (), {"quiet":True,"kill":"both"})())

dev = io.open(data_dir + "cop_scriptorium-ud-dev.conllu",encoding="utf8").read()
train = io.open(data_dir + "cop_scriptorium-ud-train.conllu",encoding="utf8").read()
train += dev
train = train.split("\n")
train = d.run_depedit(train)

test_gold = io.open(data_dir + "cop_scriptorium-ud-test.conllu",encoding="utf8").read()
# Remove supertokens and comments with empty-rule depedit
d = DepEdit(options=type('', (), {"quiet":True,"kill":"both"})())
test_gold = d.run_depedit(test_gold.split("\n"))
test_gold = test_gold.split("\n")


test_blank = []
for line in test_gold:
	if "\t" in line:
		fields = line.split("\t")
		fields[6:] = ["_","_","_","_"]
	test_blank.append(line)

test_blank = d.run_depedit(test_blank)

cmd = ["java","-jar","maltparser-1.8.jar","-c","eval","-i","tempfilename",
	   "-F",r"C:\Uni\Coptic\git\corpora\treebank-dev\v2.1parser\addMergPOSTAGS0FORMStack0.xml",
	   "-m","learn","-grl","root","-pcr","none","-a","nivrestandard","-pp","head","-nr",
	   "true","-ne","false"]

# Train the parser
exec_via_temp(train,cmd,parser_path)

# Test
cmd = ['java','-mx512m','-jar',"maltparser-1.8.jar",'-c','eval','-i','tempfilename','-m','parse']
parsed = exec_via_temp(test_blank,cmd,parser_path)
parsed = parsed.replace("\r","")

total=0
correct_head=0
correct_label=0
correct_both=0
for i, line in enumerate(parsed.split("\n")):
	gold_line = test_gold[i]
	if "\t" in line:
		total +=1
		pred = line.split("\t")
		gold = gold_line.split("\t")
		if gold[6] == pred[6] or (pred[7] == "punct" and gold[7] == "punct"):
			correct_head+=1
		if gold[7] == pred[7]:
			correct_label+=1
		if gold[6:8] == pred[6:8] or (pred[7] == "punct" and gold[7] == "punct"):
			correct_both+=1

total = float(total)
print("Label accuracy: " + str(correct_label/total))
print("Attach accuracy: " + str(correct_head/total))
print("LAS: " + str(correct_both/total))



# -*- coding: utf-8 -*-

import gc
import os
from os import listdir
from os.path import isfile, join
import re
import sys, io
from collections import defaultdict
from .xrenner_rule import CorefRule

if sys.version_info[0] < 3:
	# Python 2
	PY2 = True
	from ConfigParser import ConfigParser, NoSectionError
	def unicode_split_reader(f):
		return [line.replace("\n","").replace('\\"','"').split('\t') for line in f.read().strip().split("\n")]
		#return [line.replace("\n","").replace('\\"','"').split('\t') for line in f.readlines()]
else:
	# Python 3
	PY2 = False
	from configparser import NoSectionError, RawConfigParser as ConfigParser
	import csv

"""
LexData class - container object for lexical information, gazetteers etc.

Author: Amir Zeldes
"""

class LexData:
	"""
	Class to hold lexical information from gazetteers and training data.
	Use model argument to define subdirectory under models/ for reading different sets of
	configuration files.
	"""

	def __init__(self, model, xrenner, override=None, rule_based=False, no_seq=False):
		"""
		:param model: model - string name of the model to read from models/
		:param override: override - optional name of a section to use in models/override.ini
		:param rule_based: do not use machine learning classifiers for coreference resolution
		:param no_seq: do not use machine learning sequence taggers for entity resolution
		"""
		gc.disable()
		self.model = model
		self.docname = "untitled"
		self.atoms = {}
		self.mod_atoms = {}
		self.dump = None  # Placeholder for data dump destination
		self.dump_headers = []  # Placeholder for data dump feature names
		self.classifiers = {}  # Holds loaded classifiers from pickled files in model
		self.xrenner = xrenner
		self.entity_oracle = None  # Holds external entity predictions to use instead of system predictions
		self.oracle_counters = [0,0,0]
		self.lemma_freqs = defaultdict(float)  # Holds proportional frequency of each lemma in document
		self.token_count = 0  # Holds copy of token count for document from Xrenner object

		# Lookup model path

		if os.sep in self.model:  # Check if model provided is an absolute or relative path
			model_path = os.path.abspath(self.model)
		# Check the unlikely case that the model is directly in xrenner.py's directory
		elif os.path.exists(os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + os.sep + model):
			model_path = os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + os.sep + model
		else:  # This is a pathless model name specification, which is expected to live in ./models/
			model_path = os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + os.sep + "models" + os.sep + model

		if not os.path.exists(model_path):  # Check if this is a default model name that might need the .xrm extension
			if os.path.exists(model_path + ".xrm"):
				model_path += ".xrm"
			else:
				raise IOError("Model not found at: " + model_path)

		# Read model files and store in self.model_files dictionary

		self.model_files = {}
		if os.path.isdir(model_path):  # Check if this is a directory or a compressed model file
			model_path += os.sep
			model_files_list = [f for f in listdir(model_path) if isfile(join(model_path, f))]
			for filename in model_files_list:
				if filename.endswith(".pkl"):  # Classifier
					if not (filename.endswith("3.pkl") and PY2) and not (filename.endswith("2.pkl") and not PY2):  # Skip unpickling unneeded PY2/3 classifiers
						self.model_files[filename] = open(model_path + filename, 'rb')
				else:
					self.model_files[filename] = io.open(model_path + filename, 'r', encoding="utf8")
		else:
			from zipfile import ZipFile
			try:
				zip = ZipFile(model_path)
				model_files_list = [f for f in zip.namelist() if not os.path.isdir(f)]
				for filename in model_files_list:
					if sys.version_info[0] < 3 or filename.endswith(".pkl"):  # Python 2 or classifier
						if not (filename.endswith("3.pkl") and PY2) and not (filename.endswith("2.pkl") and not PY2):  # Skip unpickling unneeded PY2/3 classifiers
							self.model_files[filename] = zip.open(filename)
					else:
						self.model_files[filename] = io.TextIOWrapper(zip.open(filename), encoding="utf8")
			except:
				raise IOError("Could not open model file " + filename)


		self.entity_sums = defaultdict(int)

		# Get configuration
		self.filters = self.get_filters(override)
		if rule_based:
			self.filters["use_classifiers"] = False

		# Mandatory files must be included in model
		self.speaker_rules, self.non_speaker_rules = self.parse_coref_rules(self.model_files['coref_rules.tab'].read().replace("\r","").split("\n"))
		self.coref_rules = self.non_speaker_rules
		self.entities = self.read_delim(self.model_files['entities.tab'], 'quadruple') if 'entities.tab' in self.model_files else {}
		self.entity_heads = self.read_delim(self.model_files['entity_heads.tab'], 'quadruple', 'atoms', True) if 'entity_heads.tab' in self.model_files else {}
		self.pronouns = self.read_delim(self.model_files['pronouns.tab'], 'double') if "pronouns.tab" in self.model_files else {}

		# Optional files improve model accuracy
		self.names = self.read_delim(self.model_files['names.tab']) if "names.tab" in self.model_files else {}
		self.stop_list = self.read_delim(self.model_files['stop_list.tab'], 'low') if "stop_list.tab" in self.model_files else set([])
		self.open_close_punct = self.read_delim(self.model_files['open_close_punct.tab']) if "open_close_punct.tab" in self.model_files else {}
		self.open_close_punct_rev = dict((v, k) for k, v in self.open_close_punct.items())
		self.entity_mods = self.read_delim(self.model_files['entity_mods.tab'], 'triple', 'mod_atoms') if "entity_mods.tab" in self.model_files else {}
		self.entity_deps = self.read_delim(self.model_files['entity_deps.tab'], 'quadruple_numeric') if "entity_deps.tab" in self.model_files else {}
		self.lex_deps = self.read_delim(self.model_files['lex_deps.tab'], 'quadruple_numeric') if "lex_deps.tab" in self.model_files else {}
		self.hasa = self.read_delim(self.model_files['hasa.tab'], 'triple_numeric') if "hasa.tab" in self.model_files else defaultdict(lambda: defaultdict(int))
		self.coref = self.read_delim(self.model_files['coref.tab']) if "coref.tab" in self.model_files else {}
		self.numbers = self.read_delim(self.model_files['numbers.tab'], 'double') if "numbers.tab" in self.model_files else {}
		self.affix_tokens = self.read_delim(self.model_files['affix_tokens.tab']) if "affix_tokens.tab" in self.model_files else {}
		self.antonyms = self.read_antonyms() if "antonyms.tab" in self.model_files else {}
		self.isa = self.read_isa() if "isa.tab" in self.model_files else {}
		self.similar = self.read_delim(self.model_files['similar.tab'], 'double_with_sep') if "similar.tab" in self.model_files else {}
		self.nominalizations = self.read_delim(self.model_files['nominalizations.tab'], 'triple_numeric') if "nominalizations.tab" in self.model_files else {}
		self.debug = self.read_delim(self.model_files['debug.tab']) if "debug.tab" in self.model_files else {"ana":"","ante":"","ablations":""}
		self.freqs = self.read_delim(self.model_files['freqs.tab'],'double_numeric') if "freqs.tab" in self.model_files else defaultdict(int)
		additional_atoms = self.read_delim(self.model_files['atoms.tab'], 'double') if "atoms.tab" in self.model_files else {}

		# Compile atom and first + last name data
		self.atoms = self.get_atoms()
		self.atoms.update(additional_atoms)
		self.first_names, self.last_names = self.get_first_last_names(self.names)

		if self.filters["no_new_modifiers"] and self.filters["use_new_modifier_exceptions"]:
			self.exceptional_new_modifiers = self.read_delim(self.model_files['new_modifiers.tab'], 'double') if "new_modifiers.tab" in self.model_files else {}
			self.exceptional_new_modifiers.update(dict((name, 1) for name in self.first_names))
		else:
			self.exceptional_new_modifiers = {}


		self.pos_agree_mappings = self.get_pos_agree_mappings()
		self.last = {}

		self.morph = self.get_morph()
		self.func_substitutes_forward, self.func_substitutes_backward = self.get_func_substitutes()

		# Compile lemmatizer and morph rules
		self.lemma_rules = self.compile_lemmatization()
		self.morph_rules = self.compile_morph_rules()

		# Parse nested entity removal types
		self.rm_nested_entities = self.parse_rm_nested_entities()

		# Caching lists for already established non-matching pairs
		self.incompatible_mod_pairs = set([])
		self.incompatible_isa_pairs = set([])

		# Load sequence classifier if specified
		self.sequencer = None
		if "sequencer" in self.filters and not no_seq:
			if len(self.filters["sequencer"]) > 0:
				from .xrenner_sequence import Sequencer
				self.sequencer = Sequencer(model_path=self.filters["sequencer"])

		gc.enable()

	def read_delim(self, filename, mode="normal", atom_list_name="atoms", add_to_sums=False, sep=","):
		"""
		Generic file reader for lexical data in model directory

		:param filename: string - name of the file
		:param mode: double, triple, quadruple, quadruple_numeric, triple_numeric or low reading mode
		:param atom_list_name: list of atoms to use for triple reader mode
		:param add_to_sums: whether to sum numbers from multiple instances of the same key
		:param sep: separator for double_with_sep mode
		:return: compiled lexical data, usually a structured dictionary or set depending on number of columns
		"""
		if atom_list_name == "atoms":
			atom_list = self.atoms
		elif atom_list_name == "mod_atoms":
			atom_list = self.mod_atoms
		with filename as csvfile:
			if PY2:
				reader = unicode_split_reader(csvfile)
			else:
				reader = csv.reader(csvfile, delimiter='\t', escapechar="\\", quoting=csv.QUOTE_NONE)
			if mode == "low":
				return set([rows[0].lower() for rows in reader if not rows[0].startswith('#') and not len(rows[0]) == 0])
			elif mode == "double":
				out_dict = {}
				for rows in reader:
					if not rows[0].startswith('#') and not len(rows[0]) == 0:
						if rows[0] in out_dict:
							out_dict[rows[0]].append(rows[1])
						else:
							out_dict[rows[0]] = [rows[1]]
				return out_dict
			elif mode == "triple":
				out_dict = {}
				for rows in reader:
					if not rows[0].startswith('#'):
						if rows[2].endswith('@'):
							rows[2] = rows[2][0:-1]
							atom_list[rows[0]] = rows[1]
						if add_to_sums:
							self.entity_sums[rows[1]] += 1
						if rows[0] in out_dict:
							out_dict[rows[0]].append(rows[1] + "\t" + rows[2])
						else:
							out_dict[rows[0]] = [rows[1] + "\t" + rows[2]]
				return out_dict
			elif mode == "quadruple":
				out_dict = {}
				for rows in reader:
					if not rows[0].startswith('#'):
						if rows[2].endswith('@'):
							rows[2] = rows[2][0:-1]
							atom_list[rows[0]] = rows[1]
						if add_to_sums:
							self.entity_sums[rows[1]] += 1
						if len(rows) < 4:
							rows.append("0")
						if rows[0] in out_dict:
							out_dict[rows[0]].append(rows[1] + "\t" + rows[2] + "\t" + rows[3])
						else:
							out_dict[rows[0]] = [rows[1] + "\t" + rows[2] + "\t" + rows[3]]
				return out_dict
			elif mode == "double_numeric":
				out_dict = defaultdict(int)
				for row in reader:
					if not row[0].startswith("#"):
						out_dict[row[0]] = int(row[1])
				return out_dict
			elif mode == "triple_numeric":
				out_dict = defaultdict(lambda: defaultdict(int))
				for row in reader:
					if not row[0].startswith("#"):
						out_dict[row[0]][row[1]] = int(row[2])
				return out_dict
			elif mode == "quadruple_numeric":
				out_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(str)))
				for row in reader:
					if not row[0].startswith("#"):
						out_dict[row[0]][row[1]][row[2]] = int(row[3])
				return out_dict
			elif mode == "double_with_sep":
				out_dict = {}
				for row in reader:
					if not row[0].startswith("#"):
						out_dict[row[0]] = row[1].split(sep)
				return out_dict
			else:
				return dict((rows[0], rows[1]) for rows in reader if not rows[0].startswith('#'))

	def get_atoms(self):
		"""
		Function to compile atom list for atomic markable recognition. Currently treats listed persons, places,
		organizations and inanimate objects from lexical data as atomic by default.

		:return: dictionary of atoms.
		"""
		atoms = self.atoms
		places = dict((key, value[0]) for key, value in self.entities.items() if value[0].startswith(self.filters["place_def_entity"]+"\t"))
		atoms.update(places)
		atoms.update(self.names)
		persons = dict((key, value[0]) for key, value in self.entities.items() if value[0].startswith(self.filters["person_def_entity"]+"\t"))
		atoms.update(persons)
		organizations = dict((key, value[0]) for key, value in self.entities.items() if value[0].startswith(self.filters["organization_def_entity"]+"\t"))
		atoms.update(organizations)
		objects = dict((key, value[0]) for key, value in self.entities.items() if value[0].startswith(self.filters["object_def_entity"]+"\t"))
		atoms.update(objects)
		return atoms

	@staticmethod
	def get_first_last_names(names):
		"""
		Collects separate first and last name data from the collection in names.tab

		:param names: The complete names dictionary from names.tab, mapping full name to agreement
		:return: [firsts, lasts] - list containing dictionary of first names to agreement and set of last names
		"""
		firsts = {}
		lasts = set([])
		for name in names:
			if " " in name:
				parts = name.split(" ")
				firsts[parts[0]] = names[name]  # Get heuristic gender for this first name
				lasts.add(parts[len(parts)-1])  # Last name is a set, no gender info
		return [firsts,lasts]

	def read_antonyms(self):
		"""
		Function to created dictionary from each word to all its antonyms in antonyms.tab

		:return: dictionary from words to antonym sets
		"""
		set_list = self.read_delim(self.model_files['antonyms.tab'], 'low')
		output = defaultdict(set)
		for antoset in set_list:
			members = antoset.lower().split(",")
			for member in members:
				output[member].update(members)
				output[member].remove(member)
		return output

	def read_isa(self):
		"""
		Reads isa.tab into a dictionary from words to lists of isa-matches

		:return: dictionary from words to lists of corresponding isa-matches
		"""
		isa_list = self.read_delim(self.model_files['isa.tab'])
		output = {}
		for isa in isa_list:
			output[isa] = []
			members = isa_list[isa].split(",")
			for member in members:
				output[isa].append(member.lower())
		return output

	def get_filters(self, override=None):
		"""
		Reads model settings from config.ini and possibly overrides from override.ini

		:param override: optional section name in override.ini
		:return: filters - dictionary of settings from config.ini with possible overrides
		"""

		# e.g. override = 'GUM'
		config = ConfigParser()

		config.readfp(self.model_files["config.ini"])
		filters = defaultdict(str)

		# Set up default values for settings from newer versions for backwards compatibility
		filters["neg_func"] = re.compile("$^")
		filters["non_extend_pos"] = re.compile("$^")
		filters["core_infixes"] = re.compile("$^")
		filters["score_thresh"] = 0.5

		options = config.options("main")

		if override:
			config_ovrd = ConfigParser()
			config_ovrd.readfp(self.model_files['override.ini'])
			try:
				options_ovrd = config_ovrd.options(override)
			except NoSectionError:
				sys.stderr.write("\nNo section " + override + " in override.ini in model " + self.model + "\n")
				sys.exit()

		for option in options:
			if override and option in options_ovrd:
				try:
					option_string = config_ovrd.get(override, option)
					if option_string == -1:
						pass
					else:
						if option_string.startswith("/") and option_string.endswith("/"):
							option_string = option_string[1:-1]
							filters[option] = re.compile(option_string)
						elif option_string == "True" or option_string == "False":
							filters[option] = config_ovrd.getboolean(override, option)
						elif option_string.isdigit():
							filters[option] = config_ovrd.getint(override, option)
						elif option_string.count(".") == 1 and option_string.replace(".","").isdigit():
							filters[option] = config_ovrd.getfloat(override, option)
						else:
							filters[option] = option_string
				except:
					print("exception on %s!" % option)
					filters[option] = None
				continue
			try:
				option_string = config.get("main", option)
				if option_string == -1:
					pass
				else:
					if option_string.startswith("/") and option_string.endswith("/"):
						option_string = option_string[1:-1]
						filters[option] = re.compile(option_string)
					elif option_string == "True" or option_string == "False":
						filters[option] = config.getboolean("main", option)
					elif option_string.isdigit():
						filters[option] = config.getint("main", option)
					elif option_string.count(".") == 1 and option_string.replace(".","").isdigit():
						filters[option] = config.getfloat("main", option)
					else:
						filters[option] = option_string
			except AttributeError:
				print("exception on %s!" % option)
				filters[option] = None

		if ">" in filters["agree_entity_mapping"]:
			mappings = filters["agree_entity_mapping"].split(";")
			ent_map = {}
			for mapping in mappings:
				key, val = mapping.split(">")
				ent_map[key] = val
			filters["agree_entity_mapping"] = ent_map
		else:
			filters["agree_entity_mapping"] = {}

		return filters

	def compile_lemmatization(self):
		compiled_rules = []
		for rule in self.filters["lemma_rules"].split(";"):
			rule_part = rule.split("/")
			# Add a compiled lemmatization rule: regex matcher for pos, regex matcher for text, and replacement string
			compiled_rules.append((re.compile(rule_part[0]),re.compile(rule_part[1]),rule_part[2]))
		return compiled_rules

	def compile_morph_rules(self):
		compiled_rules = []
		for rule in self.filters["morph_rules"].split(";"):
			rule_part = rule.split("/")
			# Add a compiled lemmatization rule: regex matcher for pos, regex matcher for text, and replacement string
			if not rule[0] == rule[1]:
				compiled_rules.append((re.compile(rule_part[0]),rule_part[1]))
		return compiled_rules

	def lemmatize(self, token):
		"""
		Simple lemmatization function using rules from lemma_rules in config.ini

		:param token: ParsedToken object to be lemmatized
		:return: string - the lemma
		"""

		lemma = token.text
		for rule in self.lemma_rules:
			if rule[0].search(token.pos) is not None:
				lemma = rule[1].sub(rule[2], lemma)
		if self.filters["auto_lower_lemma"] == "all":
			return lemma.lower()
		elif self.filters["auto_lower_lemma"] == "except_all_caps":
			if lemma.upper() == lemma:
				return lemma
			else:
				return lemma.lower()
		else:
			return lemma

	def get_func_substitutes(self):
		"""
		Function for semi-hard-wired function substitutions based on function label and dependency direction.
		Uses func_substitute_forward and func_substitute_backward settings in config.ini

		:return: list of compiled substitutions_forward, substitutions_backward
		"""

		substitutions_forward = {}
		substitutions_backward = {}
		subst_rules = self.filters["func_substitute_forward"]
		for rule in subst_rules.split(";"):
			rule_part = rule.split("/")
			substitutions_forward[rule_part[0]] = [rule_part[1],rule_part[2]]
		subst_rules = self.filters["func_substitute_backward"]
		for rule in subst_rules.split(";"):
			rule_part = rule.split("/")
			substitutions_backward[rule_part[0]] = [rule_part[1],rule_part[2]]
		return [substitutions_forward,substitutions_backward]

	def process_morph(self, token):
		"""
		Simple mechanism for substituting values in morph feature of input tokens. For more elaborate sub-graph
		dependent manipultations, use depedit module

		:param token: ParsedToken object to edit morph feature
		:return: string - the edited morph feature
		"""

		morph = token.morph
		for rule in self.morph_rules:
			morph = rule[0].sub(rule[1], morph)
		return morph

	def get_pos_agree_mappings(self):
		"""
		Gets dictionary mapping POS categories to default agreement classes, e.g. NNS > plural

		:return: mapping dictionary
		"""

		mappings = {}
		rules = self.filters["pos_agree_mapping"]
		for rule in rules.split(";"):
			if ">" in rule:
				mappings[rule.split(">")[0]] = rule.split(">")[1]

		return mappings

	def parse_coref_rules(self,rule_list):
		"""
		Reader function to pass coref_rules.tab into CorefRule objects in two lists: one for general rules and
		one also including rules to use when speaker info is available.

		:param rule_list: textual list of rules
		:return: two separate lists of compiled CorefRule objects with and without speaker specifications
		"""

		rule_num = 0
		speaker_rules=[]
		non_speaker_rules=[]
		rule_list = [rule for rule in rule_list if len(rule) > 0 and not rule.startswith("#")]
		for rule in rule_list:
			rule_num += 1
			speaker_rules.append(CorefRule(rule, rule_num))
			if "speaker" not in rule:
				non_speaker_rules.append(CorefRule(rule, rule_num))

		# Load classifiers if available
		for rule in speaker_rules + non_speaker_rules:
			if rule.thresh is None:
				rule.thresh = self.filters["score_thresh"]
			if rule.clf_name != "_default_" and self.filters["use_classifiers"]:
				if self.filters["classifier_suffix"] != "":  # Add suffixes (e.g. for different versions in model override)
					rule.clf_name = rule.clf_name.replace(".pkl", self.filters["classifier_suffix"] + ".pkl")
				if rule.clf_name not in self.classifiers:
					rule_file_name = rule.clf_name
					if rule.clf_name not in self.model_files:  # File name missing, possible Python 2/3 variants available
						if sys.version_info[0] < 3:
							if rule.clf_name.replace(".pkl","2.pkl") in self.model_files:
								rule_file_name = rule.clf_name.replace(".pkl","2.pkl")
							else:
								if rule.clf_name.replace(".pkl", "3.pkl") in self.model_files:
									print("This model supports classifiers for Python 3 only.\n  * switch to Python 3 and try running again\n  * alternatively switch off classifiers with the option -r (expect lower accuracy)")
									sys.exit()
						elif sys.version_info[0] > 2:
							if rule.clf_name.replace(".pkl","3.pkl") in self.model_files:
								rule_file_name = rule.clf_name.replace(".pkl","3.pkl")
							else:
								if rule.clf_name.replace(".pkl", "2.pkl") in self.model_files:
									print("This model supports classifiers for Python 2 only.\n  * switch to Python 2 and try running again\n  * alternatively switch off classifiers with the option -r (expect lower accuracy)")
									sys.exit()
					try:
						from joblib import load
					except Exception as e:
						print("Unable to import joblib:\n  * classifiers in this model require installing joblib (pip install joblib)\n  * alternatively switch off classifiers with the option -r (expect lower accuracy)")
						sys.exit()
					from .xrenner_classify import Classifier

					try:
						clf = load(self.model_files[rule_file_name])
					except KeyError:
						print("\nClassifier '" + rule.clf_name + "' was not found in the model - check coref_rules.tab")
						sys.exit()
					clf = Classifier(clf[0], clf[1], clf[2])
					self.classifiers[rule.clf_name] = clf

		return speaker_rules, non_speaker_rules

	def parse_rm_nested_entities(self):
		rm_string = self.filters["remove_nested_entities"]
		types = rm_string.split(";")
		rm_nested_entities = []
		for ent_type in types:
			if ent_type.count(",") == 2:
				nested, func, container = ent_type.split(",")
				rm_nested_entities.append((nested, func, container))
		return rm_nested_entities

	def get_morph(self):
		"""
		Compiles morphlogical affix dictionary based on members of entity_heads.tab

		:return: dictionary from affixes to dictionaries mapping classes to type frequencies
		"""
		morph = {}
		for head in self.entity_heads:
			for i in range(1, self.filters["max_suffix_length"]):
				if len(head) > i:
					substring = head[len(head)-i:]
					entity_list = self.entity_heads[head]
					if substring in morph:
						for entity in entity_list:
							entity_class = entity.split("\t")[0]
							if entity_class in morph[substring]:
								morph[substring][entity_class] += 1
							else:
								morph[substring][entity_class] = 1
					else:
						for entity in entity_list:
							entity_class = entity.split("\t")[0]
							morph[substring] = {entity_class:1}
		return morph

	def read_oracle(self, oracle_file, as_text=True):

		self.entity_oracle = defaultdict(lambda : defaultdict(str))
		if not as_text:
			oracle_file = io.open(oracle_file,encoding="utf8").read()
		sents = oracle_file.strip().split("\n\n")

		for sent in sents:
			parts = sent.strip().split("\n")
			if len(parts) == 3:
				text = parts[0]
				preds = parts[-1]
				for pred in preds.split("|"):
					toks, entity = pred.split()
					start, end = toks.split(",")
					end = int(end) - 1
					start = int(start)
					self.entity_oracle[text][(start,end)] = entity





#!/usr/bin/python
# -*- coding: utf-8 -*-

import re
from collections import defaultdict, OrderedDict
from .xrenner_classes import Markable
from six import iteritems, iterkeys

"""
Marker module for markable entity recognition. Establishes compatibility between entity features
and determines markable extension in tokens

Author: Amir Zeldes
"""


def is_atomic(mark, atoms, lex):
	"""
	Checks if nested markables are allowed within this markable
	
	:param mark: the :class:`.Markable` to be checked for atomicity
	:param atoms: list of atomic markable text strings
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: bool
	"""

	marktext = mark.text.strip()

	# Do not accept a markable [New] within atomic [New Zealand]
	if marktext in atoms:
		return True
	elif marktext.lower() in atoms:
		return True
	# Remove possible prefix tokens to reject [The [United] Kingdom] if [United Kingdom] in atoms
	elif remove_prefix_tokens(marktext, lex).strip() in atoms:
		return True
	# Remove possible suffix tokens to reject [[New] Zealand 's] is [New Zealand] in atoms
	elif remove_suffix_tokens(marktext, lex).strip() in atoms:
		return True
	elif remove_infix_tokens(marktext, lex).strip() in atoms:
		return True
	# Combination of prefix and suffix to reject [The [United] Kingdom 's]
	elif mark.core_text in atoms:
		return True
	elif replace_head_with_lemma(mark) in atoms:
		return True
	# Dynamic generation of proper name pattern
	elif 0 < marktext.strip().count(" ") < 3 and marktext.strip().split(" ")[0] in lex.first_names and marktext.strip().split(" ")[-1] in lex.last_names:
		return True
	else:
		non_essential_modifiers = list(mod.text for mod in mark.head.modifiers if lex.filters["non_essential_mod_func"].match(mod.func))
		if len(non_essential_modifiers) > 0:
			mark_unmod_text = mark.core_text
			for mod in non_essential_modifiers:
				mark_unmod_text = mark_unmod_text.replace(mod+" ","")
			if mark_unmod_text in lex.atoms:
				return True
		# Not an atom, nested markables allowed
		return False


def remove_suffix_tokens(marktext, lex):
	"""
	Remove trailing tokens such as genitive 's and other tokens configured as potentially redundant to citation form

	:param marktext: the markable text string to remove tokens from
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: potentially truncated text
	"""

	if lex.filters["core_suffixes"].search(marktext):
		return lex.filters["core_suffixes"].sub(" ", marktext)
	else:
		tokens = marktext.split(" ")
		suffix_candidate = ""

		for token in reversed(tokens):
			suffix_candidate = token + " " + suffix_candidate
			if suffix_candidate.strip() in lex.affix_tokens:
				if lex.affix_tokens[suffix_candidate.strip()] == "prefix":
					return re.sub(suffix_candidate + r'$', "", marktext)
	return marktext


def remove_prefix_tokens(marktext, lex):
	"""
	Remove leading tokens such as articles and other tokens configured as potentially redundant to citation form

	:param marktext: the markable text string to remove tokens from
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: potentially truncated text
	"""

	if lex.filters["core_prefixes"].match(marktext): # NB use initial match here
		return lex.filters["core_prefixes"].sub(" ", marktext)
	else:
		tokens = marktext.split(" ")
		prefix_candidate = ""
		for token in tokens:
			prefix_candidate += token + " "
			if prefix_candidate.strip() in lex.affix_tokens:
				if lex.affix_tokens[prefix_candidate.strip()] == "prefix":
					return re.sub(r'^' + prefix_candidate, "", marktext)
	return marktext


def remove_infix_tokens(marktext, lex):
	"""
	Remove infix tokens such as dashes, interfixed articles (in Semitic construct state) etc.

	:param marktext: the markable text string to remove tokens from
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: potentially truncated text
	"""
	return lex.filters["core_infixes"].sub(" ", marktext)


def resolve_mark_entity(mark, lex):
	"""
	Main function to set entity type based on progressively less restricted parts of a markable's text

	:param mark: The :class:`.Markable` object to get the entity type for
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: void
	"""

	entity = ""
	use_entity_deps = True
	use_entity_sims = True
	use_sequencer = True if lex.sequencer is not None else False
	if "ablations" in lex.debug:
		if "no_entity_dep" in lex.debug["ablations"]:
			use_entity_deps = False
		if "no_entity_sim" in lex.debug["ablations"]:
			use_entity_sims= False
		if "no_sequencer" in lex.debug["ablations"]:
			use_sequencer = False

	## DEBUG POINT ##
	if mark.text == lex.debug["ana"] or mark.head.text == lex.debug["ana"]:
		a=5

	parent_text = mark.head.head_text
	if mark.form == "pronoun":
		if re.search(r'[12]',mark.agree):  # Explicit 1st or 2nd person pronoun
			entity = lex.filters["person_def_entity"]
			mark.entity_certainty = 'certain'
		elif mark.agree == "male" or mark.agree == "female":  # Possibly human 3rd person
			entity = lex.filters["person_def_entity"]
			mark.entity_certainty = 'uncertain'
		else:
			if use_sequencer:
				pred, score = mark.head.seq_pred
				if pred != "O":
					entity = pred
					mark.entity_certainty = 'sequencer'
			if use_entity_deps and entity == "":
				if parent_text in lex.entity_deps:
					if mark.head.func in lex.entity_deps[parent_text][mark.head.func]:
						dep_ents = dict(lex.entity_deps[parent_text][mark.head.func])
						if lex.filters["no_person_agree"].match(mark.agree) is not None and lex.filters["person_def_entity"] in dep_ents:
							del dep_ents[lex.filters["person_def_entity"]]
						if len(dep_ents) > 0:
							entity = max(iterkeys(dep_ents), key=(lambda key: dep_ents[key]))
				if entity == "":  # No literal match for dependency, fall back to similar heads
					if parent_text in lex.similar and use_entity_sims:
						similar_heads = lex.similar[parent_text]
						for similar_head in similar_heads:
							if similar_head in lex.entity_deps:
								if mark.head.func in lex.entity_deps[similar_head]:
									if lex.filters["no_person_agree"].match(mark.agree) is not None:
										similar_dict = {}
										for key, value in lex.entity_deps[similar_head][mark.head.func].items():
											if key != lex.filters["person_def_entity"]:
												similar_dict[key] = value
									else:
										similar_dict = lex.entity_deps[similar_head][mark.head.func]
									if len(similar_dict) > 0:
										entity = max(similar_dict,
													 key=(lambda key: similar_dict[key]))
										break
			if entity == "":  # Entity dependency information not used; no way to guess entity
				entity = lex.filters["default_entity"]
				mark.entity_certainty = "uncertain"
	else:
		if mark.coordinate:
			# For coordinate markables we expect the constituents to determine the entity in assign_coordinate_entity.
			# An exception to this is when the entire coordination is listed in the entities list.
			if entity == "":
				entity = resolve_entity_cascade(mark.text, mark, lex)
			if entity == "":
				entity = resolve_entity_cascade(mark.core_text, mark, lex)
		else:
			if entity == "":
				# Try to catch year numbers and hours + minutes
				if re.match(r'^(1[456789][0-9][0-9]|20[0-9][0-9]|(2[0-3]|1?[0-9]):[0-5][0-9]|ה?תש.".)$', mark.head.text) is not None:
					entity = lex.filters["time_def_entity"]
					mark.entity_certainty = "uncertain"
					mark.subclass = "time-unit" # TODO: de-hardwire this
					mark.definiteness = "def"  # literal year numbers are considered definite like 'proper names'
					mark.form = "proper"  # literal year numbers are considered definite like 'proper names'
			if entity == "":
				if re.match(r'^(([0-9]+[.,]?)+)$', mark.core_text) is not None:
					entity = lex.filters["quantity_def_entity"]
					mark.alt_entities.append(lex.filters["time_def_entity"])
					mark.entity_certainty = "uncertain"
			if entity == "":
				entity = resolve_entity_cascade(mark.text, mark, lex)
			if entity == "":
				entity = resolve_entity_cascade(replace_head_with_lemma(mark), mark, lex)
			if entity == "":
				entity = resolve_entity_cascade(remove_suffix_tokens(mark.text.strip(),lex), mark, lex)
			if entity == "":
				entity = resolve_entity_cascade(remove_prefix_tokens(mark.text.strip(), lex), mark, lex)
			if entity == "" and mark.core_text != mark.text:
				entity = resolve_entity_cascade(mark.core_text, mark, lex)
			if entity == "":
				entity = recognize_entity_by_mod(mark, lex)
			if entity == "" and mark.head.text.istitle():
				if mark.head.text in lex.last_names:
					modifiers_match_article = (lex.filters["articles"].match(mod.text) is not None for mod in mark.head.modifiers)
					modifiers_match_first_name = (mod.text in lex.first_names for mod in mark.head.modifiers)
					if any(modifiers_match_first_name) and not any(modifiers_match_article):
						entity = lex.filters["person_def_entity"]
			if entity == "" and mark.head.text.istitle():
				entity = resolve_entity_cascade(mark.core_text.lower(), mark, lex)
			if entity == "" and not mark.head.text.istitle():
				entity = resolve_entity_cascade(mark.core_text[:1].upper() + mark.core_text[1:], mark, lex)
			if entity == "":
				entity = resolve_entity_cascade(mark.head.text, mark, lex)
			if entity == "" and mark.head.text.istitle():
				entity = resolve_entity_cascade(mark.head.text.lower(), mark, lex)
			if entity == "" and mark.head.text.isupper():
				entity = resolve_entity_cascade(mark.head.text.lower(), mark, lex)
			if entity == "" and mark.head.text.isupper():
				entity = resolve_entity_cascade(mark.head.text.lower().title(), mark, lex)
			if entity == "" and not mark.head.lemma == mark.head.text:  # Try lemma match if lemma different from token
				entity = resolve_entity_cascade(mark.head.lemma, mark, lex)
			if entity == "":
				if (mark.head.text.istitle() or not lex.filters["cap_names"]):
					if mark.head.text in lex.last_names or mark.head.text in lex.first_names:
						modifiers_match_definite = (lex.filters["definite_articles"].match(mod.text) is not None for mod in mark.head.modifiers)
						modifiers_match_article = (lex.filters["articles"].match(mod.text) is not None for mod in mark.head.modifiers)
						modifiers_match_def_entity = (re.sub(r"\t.*","",lex.entity_heads[mod.text.strip().lower()][0]) == lex.filters["default_entity"] for mod in mark.head.modifiers if mod.text.strip().lower() in lex.entity_heads)
						if not (any(modifiers_match_article) or any(modifiers_match_definite) or any(modifiers_match_def_entity)):
							entity = lex.filters["person_def_entity"]
			if entity == "":
				# Just use sequencer if desired
				if use_sequencer:
					pred, score = mark.head.seq_pred
					if pred != "O":
						entity = pred
						mark.entity_certainty = 'sequencer'

			if entity == "":
				# See what the affix morphology predicts for the head
				head_text = mark.lemma if mark.lemma != "_" and mark.lemma != "" else mark.head.text
				morph_probs = get_entity_by_affix(head_text,lex)

				# Now check what the dependencies predict
				dep_probs = {}
				if use_entity_deps:
					if parent_text in lex.entity_deps:
						if mark.head.func in lex.entity_deps[parent_text]:
							dep_probs.update(lex.entity_deps[parent_text][mark.head.func])
					if len(dep_probs) == 0:  # No literal dependency information found, check if similar heads are known
						if parent_text in lex.similar:
							similar_heads = lex.similar[parent_text]
							for similar_head in similar_heads:
								if similar_head in lex.entity_deps:
									if mark.head.func in lex.entity_deps[similar_head]:
										dep_probs.update(lex.entity_deps[similar_head][mark.head.func])
										break

				# And check what entity similar words are
				sim_probs = {}
				if use_entity_sims:
					if mark.head.text in lex.similar:
						for similar_word in lex.similar[mark.head.text]:
							if similar_word in lex.entity_heads:
								for entity_type in lex.entity_heads[similar_word]:
									entity_string = entity_type.split("\t")[0]
									if entity_string in sim_probs:
										sim_probs[entity_string] += 1
									else:
										sim_probs.update({entity_string:1})

				# Compare scores to decide between affix vs. dependency evidence vs. embeddings
				dep_values = list(dep_probs[key] for key in dep_probs)
				total_deps = float(sum(dep_values))
				sim_values = list(sim_probs[key] for key in sim_probs)
				total_sims = float(sum(sim_values))
				norm_dep_probs = {}
				norm_sim_probs = {}

				# Normalize - each information source hedges its bets based on how many guesses it makes
				for key, value in iteritems(dep_probs):
					norm_dep_probs[key] = value/total_deps
				for key, value in iteritems(sim_probs):
					norm_sim_probs[key] = value/total_sims

				joint_probs = defaultdict(float)
				joint_probs.update(norm_dep_probs)
				for entity in morph_probs:
					joint_probs[entity] += morph_probs[entity]
				for entity in norm_sim_probs:
					joint_probs[entity] += sim_probs[entity]
				# Bias in favor of default entity to break ties
				joint_probs[lex.filters["default_entity"]] += 0.0000001

				entity = max(joint_probs, key=(lambda key: joint_probs[key]))

	if entity != "":
		mark.entity = entity

	if "/" in mark.entity:  # Lexicalized agreement information appended to entity
		if mark.agree == "" or mark.agree is None:
			mark.agree = mark.entity.split("/")[1]
		elif mark.agree_certainty == "":
			mark.alt_agree.append(mark.agree)
			mark.agree = mark.entity.split("/")[1]
		mark.entity = mark.entity.split("/")[0]
	elif mark.entity == lex.filters["person_def_entity"] and mark.agree == lex.filters["default_agree"] and mark.form != "pronoun":
		mark.agree = lex.filters["person_def_agree"]
		mark.agree_certainty = "uncertain"
	if "\t" in mark.entity:  # This is a subclass bearing solution
		mark.subclass = mark.entity.split("\t")[1]
		mark.entity = mark.entity.split("\t")[0]
	if mark.entity == lex.filters["person_def_entity"] and mark.form != "pronoun":
		if mark.text in lex.names:
			mark.agree = lex.names[mark.text]
	if mark.entity == lex.filters["person_def_entity"] and mark.agree is None:
		no_affix_mark = remove_suffix_tokens(remove_prefix_tokens(mark.text, lex), lex)
		if no_affix_mark in lex.names:
			mark.agree = lex.names[no_affix_mark]
	if mark.entity == lex.filters["person_def_entity"] and mark.agree is None:
		mark.agree = lex.filters["person_def_agree"]
		mark.agree_certainty = "uncertain"
	if mark.entity == "" and mark.core_text.upper() == mark.core_text and re.search(r"[A-ZÄÖÜ]", mark.core_text) is not None:  # Unknown all caps entity, guess acronym default
		mark.entity = lex.filters["all_caps_entity"]
		mark.entity_certainty = "uncertain"
	if mark.entity == "":  # Unknown entity, guess default
		mark.entity = lex.filters["default_entity"]
		mark.entity_certainty = "uncertain"
	if mark.subclass == "":
		if mark.subclass == "":
			mark.subclass = mark.entity
	if mark.func == "title":
		mark.entity = lex.filters["default_entity"]
	if mark.agree == "" and mark.entity == lex.filters["default_entity"]:
		mark.agree = lex.filters["default_agree"]


def resolve_entity_cascade(entity_text, mark, lex):
	"""
	Retrieve possible entity types for a given text fragment based on entities list, entity heads and names list.

	:param entity_text: The text to determine the entity for
	:param mark: The :class:`.Markable` hosting the text fragment to retrieve context information from (e.g. dependency)
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: entity type; note that this is used to decide whether to stop the search, but the Markable's entity is	already set during processing together with matching subclass and agree information
	"""
	options = {}
	entity = ""
	person_entity = lex.filters["person_def_entity"]
	if entity_text in lex.entities:
		for alt in lex.entities[entity_text]:
			parsed_entity = parse_entity(alt, "entities_match")
			if parsed_entity[0] not in mark.alt_entities:
				mark.alt_entities.append(parsed_entity[0])
				mark.alt_subclasses.append(parsed_entity[1])
				options[parsed_entity[0]] = parsed_entity
	if entity_text in lex.entity_heads:
		for alt in lex.entity_heads[entity_text]:
			parsed_entity = parse_entity(alt, "entity_heads_match")
			if parsed_entity[0] not in mark.alt_entities:
				mark.alt_entities.append(parsed_entity[0])
				mark.alt_subclasses.append(parsed_entity[1])
				options[parsed_entity[0]] = parsed_entity
	# Add the person entity based on a possible name despite seeing alternative entities
	# If and only if this is supported by a unique dependency cue (only person dependencies found, well attested)
	if entity_text in lex.names or entity_text in lex.last_names or entity_text in lex.first_names:
		if (entity_text[0].istitle() or not lex.filters["cap_names"]) and person_entity not in mark.alt_entities:
			if mark.head.head_text in lex.entity_deps:
				if mark.func in lex.entity_deps[mark.head.head_text]:
					if lex.filters["person_def_entity"] in lex.entity_deps[mark.head.head_text][mark.func]:
						# Must be attested > 5 times; relaxing this can lower precision substantially
						if lex.entity_deps[mark.head.head_text][mark.func][lex.filters["person_def_entity"]] > 5 and len(lex.entity_deps[mark.head.head_text][mark.func])==1:
							mark.alt_entities.append(lex.filters["person_def_entity"])
							mark.alt_subclasses.append(lex.filters["person_def_entity"])
							name_agree = ""
							if entity_text in lex.names:
								name_agree = lex.names[entity_text]
							elif entity_text in lex.first_names and not entity_text in lex.last_names:
								name_agree = lex.first_names[entity_text]
							options[person_entity] = (person_entity, person_entity, name_agree,"names_match")
	if len(mark.alt_entities) < 1 and 0 < entity_text.count(" ") < 3 and lex.filters["person_def_entity"] not in mark.alt_entities:
		if entity_text.split(" ")[0] in lex.first_names and entity_text.split(" ")[-1] in lex.last_names:
			if entity_text[0].istitle() or not lex.filters["cap_names"]:
				if lex.filters["articles"].match(mark.text.split(" ")[0]) is None:
					mark.alt_entities.append(person_entity)
					mark.alt_subclasses.append(person_entity)
					options[person_entity] = (person_entity, person_entity, lex.first_names[entity_text.split(" ")[0]], "name_match")

	if person_entity not in mark.alt_entities and (mark.text in lex.first_names or mark.text in lex.last_names):
		mark.alt_entities.append(person_entity)
		options[person_entity] = (person_entity,person_entity,'','name_match')
	if len(mark.alt_entities) > 1:
		entity = disambiguate_entity(mark, lex)
	elif len(mark.alt_entities) == 1:
		entity = mark.alt_entities[0]

	if entity != "":
		mark.entity, mark.subclass = options[entity][0:2]
		if options[entity][2] != "":
			mark.agree = options[entity][2]
		mark.entity_certainty = options[entity][3]

	return entity if len(options) > 0 else ""


def parse_entity(entity_text, certainty="uncertain"):
	"""
	Parses: entity -tab- subclass(/agree) + certainty into a tuple

	:param entity_text: the string to parse, must contain excatly two tabs
	:param certainty: the certainty string at end of tuple, default 'uncertain'
	:return: quadruple of (entity, subclass, agree, certainty)
	"""
	entity, subclass, freq = entity_text.split("\t")
	if "/" in subclass:
		subclass, agree = subclass.split("/")
	else:
		agree = ""
	return (entity, subclass, agree, certainty)


def resolve_mark_agree(mark, lex):
	"""
	Resolve Markable agreement based on morph information in tokens or gazetteer data

	:param mark: The :class:`.Markable` to resolve agreement for
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: void
	"""

	# DEBUG POINT #
	if mark.text == lex.debug["ana"]:
		a=5

	if mark.head.morph not in ["","_"]:
		mark.agree_certainty = "head_morph"
		return [mark.head.morph]
	else:
		if mark.form == "pronoun":
			if mark.text in lex.pronouns:
				return lex.pronouns[mark.text]
			elif mark.text.lower() in lex.pronouns:
				return lex.pronouns[mark.text.lower()]
		if mark.form == "proper":
			if mark.core_text in lex.names:
				return [lex.names[mark.core_text]]
			elif mark.core_text in lex.first_names and mark.core_text not in lex.entities and mark.core_text not in lex.entity_heads:  # Single name component core text
				return [lex.first_names[mark.core_text]]
		if mark.head.pos in lex.pos_agree_mappings:
			mark.agree_certainty = "pos_agree_mappings"
			return [lex.pos_agree_mappings[mark.head.pos]]
		elif mark.core_text in lex.entities:
			for full_entry in lex.entities[mark.core_text]:
				entry = full_entry.split("\t")[1]
				if "/" in entry:
					if mark.agree == "":
						mark.agree = entry[entry.find("/") + 1:]
					mark.alt_agree.append(entry[entry.find("/") + 1:])
		elif mark.head.text in lex.entity_heads:
			for full_entry in lex.entity_heads[mark.head.text]:
				entry = full_entry.split("\t")[1]
				if "/" in entry:
					if mark.agree == "":
						mark.agree = entry[entry.find("/") + 1:]
					mark.alt_agree.append(entry[entry.find("/") + 1:])


def resolve_cardinality(mark,lex):
	"""
	Find cardinality for Markable based on numerical modifiers or number words

	:param mark: The :class:`.Markable` to resolve agreement for
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: Cardinality as float, zero if unknown
	"""
	def check_card(mod):
		if mod in lex.numbers:
			return int(lex.numbers[mod][0])
		elif mod.lower() in lex.numbers:
			return int(lex.numbers[mod.lower()][0])
		else:
			thousand_sep = r"\." if lex.filters["thousand_sep"] == "." else lex.filters["thousand_sep"]
			pure_number_candidate = re.sub(thousand_sep,"",mod)

			decimal_sep = lex.filters["decimal_sep"]
			if decimal_sep != ".":
				pure_number_candidate = re.sub(decimal_sep,".",pure_number_candidate)
			if re.match("^(\d+(\.\d+)?|(\.\d+))$",pure_number_candidate) is not None:
				return float(pure_number_candidate)
			else:
				parts = re.match("^(\d+)/(\d+)$",pure_number_candidate)
				if parts is not None: # Fraction with slash division
					numerator = float(parts.groups()[0])
					denominator = float(parts.groups()[1])
					return numerator/denominator

	for m in mark.head.modifiers:
		card = check_card(m.text)
		if card is not None:
			return card
	card = check_card(mark.head.text)
	if card is not None:
		return card
	card = check_card(mark.head.lemma)
	if card is not None:
		return card

	return 0


def recognize_entity_by_mod(mark, lex, mark_atoms=False):
	"""
	Attempt to recognize entity type based on modifiers
	
	:param mark: :class:`.Markable` for which to identify the entity type
	:param modifier_lexicon: The :class:`.LexData` object's modifier list
	:return: String (entity type, possibly including subtype and agreement)
	"""
	modifier_lexicon = lex.entity_mods
	mod_atoms = lex.mod_atoms
	for mod in mark.head.modifiers:
		modifier_tokens = [mod.text] + [construct_modifier_substring(mod)]
		while len(modifier_tokens) > 0:
			identifying_substr = ""
			for token in modifier_tokens:
				identifying_substr += token + " "
				if identifying_substr.strip() in modifier_lexicon:
					if identifying_substr.strip() in mod_atoms and mark_atoms:
						return modifier_lexicon[identifying_substr.strip()][0] + "@"
					else:
						return modifier_lexicon[identifying_substr.strip()][0]
				elif identifying_substr.lower().strip() in modifier_lexicon:
					if identifying_substr.lower().strip() in mod_atoms and mark_atoms:
						return modifier_lexicon[identifying_substr.lower().strip()][0] + "@"
					else:
						return modifier_lexicon[identifying_substr.lower().strip()][0]
			modifier_tokens.pop(0)
	return ""


def construct_modifier_substring(modifier):
	"""
	Creates a list of tokens representing a modifier and all of its submodifiers in sequence
	
	:param modifier: A ParsedToken object from the modifier list of the head of some markable
	:return: Text of that modifier together with its modifiers in sequence
	"""
	candidate_prefix = ""
	mod_dict = get_mod_ordered_dict(modifier)
	for mod_member in mod_dict:
		candidate_prefix += mod_dict[mod_member].text + " "
	return candidate_prefix.strip()


def stoplist_prefix_tokens(mark, prefix_dict, keys_to_pop):
	substr = ""
	candidate_prefix = ""
	for mod in mark.head.modifiers:
		mod_dict = get_mod_ordered_dict(mod)
		for mod_member in mod_dict:
			candidate_prefix += mod_dict[mod_member].text + " "
		tokens = candidate_prefix.strip().split(" ")
		for token in tokens:
			substr += token + " "
			if substr.strip() in prefix_dict:
				tokens_affected_count = substr.count(" ")
				i = 0
				for mod_token in mod_dict:
					if i < tokens_affected_count and not mod_dict[mod_token].id == mark.head.id:
						keys_to_pop.append(mod_dict[mod_token].id)
					i += 1


def get_mod_ordered_dict(mod):
	"""
	Retrieves the (sub)modifiers of a modifier token
	
	:param mod: A :class:`.ParsedToken` object representing a modifier of the head of some markable
	:return: Recursive ordered dictionary of that modifier's own modifiers
	"""
	mod_dict = OrderedDict()
	mod_dict[int(mod.id)] = mod
	if len(mod.modifiers) > 0:
		for mod2 in mod.modifiers:
			mod_dict.update(get_mod_ordered_dict(mod2))
	else:
		return mod_dict
	return OrderedDict(sorted(mod_dict.items()))


def markable_extend_punctuation(marktext, adjacent_token, punct_dict, direction):
	if direction == "trailing":
		for open_punct in punct_dict:
			if (" " + open_punct + " " in marktext or marktext.startswith(open_punct + " ")) and adjacent_token.text == punct_dict[open_punct]:
				return True
	else:
		for close_punct in punct_dict:
			if (" " + close_punct + " " in marktext or marktext.endswith(" " + close_punct)) and adjacent_token.text == punct_dict[close_punct]:
				return True
	return False


def markables_overlap(mark1, mark2, lex=None):
	"""
	Helper function to check if two markables cover some of the same tokens. Note that if the lex argument is specified,
	it is used to recognize possessives, which behave exceptionally. Possessive pronouns beginning
	after a main markable has started are tolerated in case of markable definitions including relative clauses,
	e.g. [Mr. Pickwick, who was looking for [his] hat]

	:param mark1: First :class:`.Markable`
	:param mark2: Second :class:`.Markable`
	:param lex: the :class:`.LexData` object with gazetteer information and model settings or None
	:return: bool
	"""
	if lex is not None:
		if lex.filters["possessive_func"].match(mark1.func) is not None and mark1.form == "pronoun" and mark1.start > mark2.start:
			return False
		elif lex.filters["possessive_func"].match(mark2.func) is not None and mark2.form == "pronoun" and mark2.start > mark1.start:
			return False
	if mark2.end >= mark1.start >= mark2.start:
		return True
	elif mark2.end >= mark1.end >= mark2.start:
		return True
	else:
		return False


def markable_extend_affixes(start, end, conll_tokens, sent_start, lex):
	candidate_affix = ""
	for tok in reversed(conll_tokens[sent_start:start]):
		candidate_affix = tok.text + " " + candidate_affix
		if candidate_affix.lower().strip() in lex.affix_tokens:
			if lex.affix_tokens[candidate_affix.lower().strip()] == "prefix":
				return [int(tok.id), int(tok.id) + candidate_affix.count(" ")]
		elif candidate_affix.strip() in lex.affix_tokens:
			if lex.affix_tokens[candidate_affix.strip()] == "prefix":
				return [int(tok.id), int(tok.id) + candidate_affix.count(" ")]
	candidate_affix = ""
	for tok in conll_tokens[end+1:]:
		candidate_affix += tok.text + " "
		if candidate_affix.lower().strip() in lex.affix_tokens:
			if lex.affix_tokens[candidate_affix.lower().strip()] == "suffix":
				return [int(tok.id) - candidate_affix.strip().count(" "), int(tok.id) + 1]
		elif candidate_affix.strip() in lex.affix_tokens:
			if lex.affix_tokens[candidate_affix.strip()] == "suffix":
				return [int(tok.id) - candidate_affix.strip().count(" "), int(tok.id) + 1]
	return [0,0]


def get_entity_by_affix(head_text, lex):
	affix_max = 0
	score = 0
	entity = ""
	probs = {}
	for i in range(1, len(head_text) - 1):
		candidates = 0
		if head_text[i:] in lex.morph:
			options = lex.morph[head_text[i:]]
			for key, value in options.items():
				candidates += value
				entity = key.split("/")[0]
				probs[entity] = float(value)
			for entity in probs:
				probs[entity] = probs[entity] / candidates
		if entity != "":
			return probs
	return probs


def pos_func_combo(pos, func, pos_func_heads_string):
	"""
	:return: bool
	"""
	pos_func_heads = pos_func_heads_string.split(";")
	if pos + "+" + func in pos_func_heads:
		return True
	elif pos + "!" + func in pos_func_heads:
		return False
	else:
		if pos_func_heads_string.find(";" + pos + "!") > -1 or pos_func_heads_string.startswith(pos + "!"):
			return True
		else:
			return False


def replace_head_with_lemma(mark):
	head = re.escape(mark.head.text)
	lemma = mark.head.lemma
	return re.sub(head, lemma, mark.core_text).strip()


def make_markable(tok, conll_tokens, descendants, tokoffset, sentence, keys_to_pop, lex):
	if tok.id in descendants and lex.filters["non_extend_pos"].match(tok.pos) is None:
		tokenspan = descendants[tok.id] + [tok.id]
		tokenspan = list(map(int, tokenspan))
		tokenspan.sort()
		marktext = ""
		start = min(tokenspan)
		end = max(tokenspan)
		for span_token in conll_tokens[start:end + 1]:
			marktext += span_token.text + " "
		marktext = marktext.strip()
	else:
		marktext = tok.text
		start = int(tok.id)
		end = int(tok.id)
	# Check for a trailing coordinating conjunction on a descendant of the head and re-connect if necessary
	if end < len(conll_tokens) - 1:
		coord = conll_tokens[end + 1]
		if lex.filters["cc_left_to_right"]:
			not_head_child = coord.head != tok.id
		else:
			coord_grand_head = 0
			if int(conll_tokens[int(coord.head)].head) != 0:
				coord_grand_head = int(conll_tokens[int(coord.head)].head)
			not_head_child = (conll_tokens[int(coord.head)].head != tok.id
									and coord_grand_head == int(tok.id)
									and conll_tokens[int(coord.head)].head != '0'
									and int(conll_tokens[int(coord.head)].head) > int(tok.id))

		if lex.filters["coord_func"].match(coord.func) is not None and not_head_child and int(coord.head) >= start:
			conjunct1 = conll_tokens[int(conll_tokens[end + 1].head)]
			for tok2 in conll_tokens[end + 1:]:
				if (tok2.head == conjunct1.head and tok2.func == conjunct1.func) or tok2.head == coord.id:
					conjunct2 = tok2
					tokenspan = [conjunct2.id, str(end)]
					if conjunct2.id in descendants:
						tokenspan += descendants[conjunct2.id]
					tokenspan = map(int, tokenspan)
					tokenspan = sorted(tokenspan)
					end = max(tokenspan)
					marktext = ""
					for span_token in conll_tokens[start:end + 1]:
						marktext += span_token.text + " "
					break

	core_text = marktext.strip()
	# DEBUG POINT
	if marktext.strip() in lex.debug:
		pass
	# Extend markable to 'affix tokens'
	# Do not extend pronouns or stop functions
	if lex.filters["stop_func"].match(tok.func) is None and lex.filters["pronoun_pos"].match(tok.pos) is None:
		extend_affixes = markable_extend_affixes(start, end, conll_tokens, tokoffset + 1, lex)
		if not extend_affixes[0] == 0:
			if extend_affixes[0] < start:
				prefix_text = ""
				for prefix_tok in conll_tokens[extend_affixes[0]:extend_affixes[1]]:
					prefix_text += prefix_tok.text + " "
					keys_to_pop.append(prefix_tok.id)
					start -= 1
				marktext = prefix_text + marktext
			else:
				for suffix_tok in conll_tokens[extend_affixes[0]:extend_affixes[1]]:
					keys_to_pop.append(suffix_tok.id)
					marktext += suffix_tok.text + " "
					end += 1

	# Extend markable to trailing closing punctuation if it contains opening punctuation
	if end < len(conll_tokens) - 1:
		next_id = end + 1
		if markable_extend_punctuation(marktext, conll_tokens[next_id], lex.open_close_punct, "trailing"):
			marktext += conll_tokens[next_id].text + " "
			end += 1
	if start != "1":
		prev_id = start - 1
		if markable_extend_punctuation(marktext, conll_tokens[prev_id], lex.open_close_punct_rev, "leading"):
			marktext = conll_tokens[prev_id].text + " " + marktext
			start -= 1

	this_markable = Markable("", tok, "", "", start, end, core_text, core_text, "", "", "", "new", "", sentence, "none", "none", 0, [], [], [])
	# DEBUG POINT
	if this_markable.text == lex.debug["ana"]:
		pass
	this_markable.core_text = remove_infix_tokens(remove_suffix_tokens(remove_prefix_tokens(this_markable.core_text,lex),lex),lex)
	while this_markable.core_text != core_text:  # Potentially repeat affix stripping as long as core text changes
		core_text = this_markable.core_text
		this_markable.core_text = remove_infix_tokens(remove_suffix_tokens(remove_prefix_tokens(this_markable.core_text,lex),lex),lex)
	if this_markable.core_text == '':  # Check in case suffix removal has left no core_text
		this_markable.core_text = marktext.strip()
	this_markable.text = marktext  # Update core_text with potentially modified markable text
	return this_markable


def lookup_has_entity(text, lemma, entity, lex):
	"""
	Checks if a certain token text or lemma have the specific entity listed in the entities or entity_heads lists
	
	:param text: text of the token
	:param lemma: lemma of the token
	:param entity: entity to check for
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: bool
	"""
	found = []
	if text in lex.entities:
		found = [i for i, x in enumerate(lex.entities[text]) if re.search(entity + '\t', x)]
	elif lemma in lex.entities:
		found = [i for i, x in enumerate(lex.entities[lemma]) if re.search(entity + '\t', x)]
	elif text in lex.entity_heads:
		found = [i for i, x in enumerate(lex.entity_heads[text]) if re.search(entity + '\t', x)]
	elif lemma in lex.entity_heads:
		found = [i for i, x in enumerate(lex.entity_heads[lemma]) if re.search(entity + '\t', x)]
	return len(found) > 0


def assign_coordinate_entity(mark,markables_by_head):
	"""
	Checks if all constituents of a coordinate markable have the same entity and subclass and if so, propagates these to the coordinate markable.

	:param mark: a coordinate markable to check the entities of its constituents
	:param markables_by_head: dictionary of markables by head id
	:return: void
	"""

	sub_entities = []
	sub_subclasses = []
	for m_id in mark.submarks:
		if m_id in markables_by_head:
			sub_entities.append(markables_by_head[m_id].entity)
			sub_subclasses.append(markables_by_head[m_id].subclass)
	if len(set(sub_entities)) == 1:  # There is agreement on the entity
		mark.entity = sub_entities[0]
	if len(set(sub_subclasses)) == 1:  # There is agreement on the entity
		mark.subclass = sub_subclasses[0]


def disambiguate_entity(mark,lex):
	"""
	Selects prefered entity for a Markable with multiple alt_entities based on dependency information or more common type

	:param mark: the Markable object
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: predicted entity type as string
	"""
	##DEBUG POINT##
	if mark.text in lex.debug["ana"]:
		a=3

	# Prefer sequencer entity if it is one of the options
	use_sequencer = True if lex.sequencer is not None else False
	if use_sequencer:
		seq_ent = mark.head.seq_pred[0]
		if seq_ent in mark.alt_entities:
			return seq_ent

	parent_text = mark.head.head_text
	scores = defaultdict(float)
	entity_freqs = defaultdict(int)
	# Bias tie breaker in favor of default entity
	if lex.filters["default_entity"] in mark.alt_entities:
		scores[lex.filters["default_entity"]] += 0.0001
	if parent_text in lex.entity_deps:
		if mark.func in lex.entity_deps[parent_text]:
			for alt_entity in mark.alt_entities:
				if alt_entity in lex.entity_deps[parent_text][mark.func]:
					entity_freqs[alt_entity] = lex.entity_deps[parent_text][mark.func][alt_entity]

	if len(entity_freqs) == 0:  # No dependency info, use similar dependencies if available
		if parent_text in lex.similar:
			for similar_parent in lex.similar[parent_text]:
				if similar_parent in lex.entity_deps:
					if mark.func in lex.entity_deps[similar_parent]:
						for alt_entity in mark.alt_entities:
							if alt_entity in lex.entity_deps[similar_parent][mark.func]:
								entity_freqs[alt_entity] = lex.entity_deps[similar_parent][mark.func][alt_entity]

	# Check for ties
	break_tie = False
	if len(entity_freqs) > 0:
		best_freq = max(entity_freqs.values())
		best_entities = [k for k, v in entity_freqs.items() if v == best_freq]
		if len(best_entities) > 1:
			break_tie = True

	if len(entity_freqs) == 0 or break_tie:  # No similar dependencies, get frequency information from entities if available
		if mark.text in lex.entities:
			for entity_entry in lex.entities[mark.text]:
				entity_type, entity_subtype, freq = entity_entry.split("\t")
				freq = int(freq)
				if freq > 0:
					entity_freqs[entity_type] += freq
	if len(entity_freqs) == 0 or break_tie:  # No similar dependencies, get frequency information from heads if available
		if mark.head.text in lex.entity_heads:
			for entity_entry in lex.entity_heads[mark.head.text]:
				entity_type, entity_subtype, freq = entity_entry.split("\t")
				freq = int(freq)
				if freq > 0:
					entity_freqs[entity_type] += freq

	if len(entity_freqs) == 0:  # No dependency info, use entity sum proportions
		entity_freqs = lex.entity_sums

	for entity_type in mark.alt_entities:
		scores[entity_type] += entity_freqs[entity_type]


	best_entity = max(iterkeys(scores), key=(lambda key: scores[key]))
	return best_entity




#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
DepEdit - A simple configurable tool for manipulating dependency trees

Input: CoNLL10 or CoNLLU (10 columns, tab-delimited, blank line between sentences, comments with pound sign #)

Author: Amir Zeldes
"""

from __future__ import print_function
import argparse
import re
from io import open as io_open
from copy import copy, deepcopy
import sys, os
from collections import defaultdict
from glob import glob
from six import iteritems

__version__ = "2.1.2"


def escape(string,symbol_to_mask,border_marker):
	inside = False
	output = ""
	# Scan the string looking for border or symbol to mask
	for char in string:
		if char == border_marker:
			inside = not inside
		if char == symbol_to_mask and inside:
			output += "%%%%%"
		else:
			output += char
	return output


class ParsedToken:
	def __init__(self, tok_id, text, lemma, pos, cpos, morph, head, func, head2, func2, num, child_funcs, position, is_super_tok=False):
		self.id = tok_id
		self.text = text
		self.pos = pos
		self.cpos = cpos
		self.lemma = lemma
		self.morph = morph
		self.head = head
		self.func = func
		self.head2 = head2
		self.func2 = func2
		self.num = num
		self.child_funcs = child_funcs
		self.position = position
		self.is_super_tok = is_super_tok

	def __repr__(self):
		return str(self.text) + " (" + str(self.pos) + "/" + str(self.lemma) + ") " + "<-" + str(self.func)


class Sentence:

	def __init__(self,sentence_string=""):
		self.sentence_string = sentence_string
		self.length = 0
		self.annotations = {}

	def print_annos(self):
		out_string = ""
		for (key,val) in iteritems(self.annotations):
			out_string += "# " + key + "=" + val + "\n"
		return out_string


class Transformation:

	def parse_transformation(self, transformation_text):
		if transformation_text.count("\t") < 2:
			return None
		else:
			split_trans = transformation_text.split("\t")
			definition_string = split_trans[0]
			relation_string = split_trans[1]
			action_string = split_trans[2]
			relation_string = self.normalize_shorthand(relation_string)
			action_string = self.normalize_shorthand(action_string)
			definition_string = escape(definition_string,";","/")
			definitions_list = definition_string.split(";")
			escaped_definitions = []
			for _def in definitions_list:
				escaped_definitions.append(_def.replace("%%%%%",";"))
			definitions = []
			for def_index, esc_string in enumerate(escaped_definitions):
				definitions.append(DefinitionMatcher(esc_string,def_index + 1))
			relations = relation_string.split(";")
			actions = action_string.strip().split(";")
			aliased_actions = []
			for action in actions:
				aliased_actions.append(self.handle_aliases(action))
			return [definitions, relations, aliased_actions]

	@staticmethod
	def handle_aliases(orig_action):
		orig_action = orig_action.replace(":form=",":text=")
		orig_action = orig_action.replace(":upostag=", ":pos=")
		orig_action = orig_action.replace(":xpostag=",":cpos=")
		orig_action = orig_action.replace(":feats=",":morph=")
		orig_action = orig_action.replace(":deprel=",":func=")
		orig_action = orig_action.replace(":deps=",":head2=")
		orig_action = orig_action.replace(":misc=",":func2=")
		return orig_action

	@staticmethod
	def normalize_shorthand(criterion_string):
		criterion_string = criterion_string.replace('.*', '.1,1000')
		temp = ""
		while temp != criterion_string:
			temp = criterion_string
			criterion_string = re.sub(r'(#[0-9]+)(>|\.(?:[0-9]+(?:,[0-9]+)?)?)(#[0-9]+)(>|\.(?:[0-9]+(?:,[0-9]+)?)?)', r'\1\2\3;\3\4', criterion_string)
		return criterion_string

	def __init__(self, transformation_text, line):
		instructions = self.parse_transformation(transformation_text)
		if instructions is None:
			message = "Depedit says: error in configuration file\n"
			message += "Malformed instruction on line " + str(line) + " (instruction lines must contain exactly two tabs)\n"
			sys.stderr.write(message)
			sys.exit()
		self.definitions = instructions[0]
		self.relations = instructions[1]
		self.actions = instructions[2]
		self.line = line

	def validate(self):
		report = ""
		for definition in self.definitions:
			node = definition.def_text
			node = escape(node,"&","/")
			criteria = node.split("&")
			criteria = (_crit.replace("%%%%%","&") for _crit in criteria)
			for criterion in criteria:
				if re.match("(text|pos|cpos|lemma|morph|func|head|func2|head2|num|form|upostag|xpostag|feats|deprel|deps|misc)!?=/[^/=]*/",criterion) is None:
					if re.match(r"position!?=/(first|last|mid)/",criterion) is None:
						report+= "Invalid node definition in column 1: " + criterion
		for relation in self.relations:
			if relation == "none" and len(self.relations) == 1:
				if len(self.definitions) > 1:
					report += "Column 2 setting 'none' invalid with more than one definition in column 1"
			elif relation == "none":
				report += "Setting 'none' invalid in column 2 when multiple relations are defined"
			else:
				criteria = relation.split(";")
				for criterion in criteria:
					criterion = criterion.strip()
					#if not re.match(r"#[0-9]+((>|\.([0-9]+(,[0-9]+)?)?)#[0-9]+)+",criterion):
					if not re.match(r"(#[0-9]+((>|\.([0-9]+(,[0-9]+)?)?)#[0-9]+)+|#[0-9]+:(text|pos|cpos|lemma|morph|func|head|func2|head2|num|form|upostag|xpostag|feats|deprel|deps|misc)==#[0-9]+)", criterion):
						report += "Column 2 relation setting invalid criterion: " + criterion + "."
		for action in self.actions:
			commands = action.split(";")
			for command in commands:
				if not re.match(r"(#[0-9]+>#[0-9]+|#[0-9]+:(func|lemma|text|pos|cpos|morph|head|head2|func2|num|form|upostag|xpostag|feats|deprel|deps|misc)=[^;]*)$",command):  # Node action
					if not re.match(r"#S:[A-Za-z_]+=[A-Za-z_]+$|last$", command):  # Sentence annotation action or quit
						report += "Column 3 invalid action definition: " + command + " and the action was " + action
		return report


class DefinitionMatcher:

	def __init__(self, def_text, def_index):
		self.def_text = escape(def_text,"&","/")
		self.def_index = def_index
		self.groups = []
		self.defs = []

		def_items = self.def_text.split("&")
		for def_item in def_items:
			def_item = def_item.replace("%%%%%","&")
			criterion = def_item.split("=",1)[0]
			if criterion[-1] == "!":
				negative_criterion = True
				criterion = criterion[:-1]
			else:
				negative_criterion = False

			def_value = def_item.split("=",1)[1][1:-1]

			# Ensure regex is anchored
			if def_value[0] != "^":
				def_value = "^" + def_value
			if def_value[-1] != "$":
				def_value += "$"
			self.defs.append(Definition(criterion,def_value,negative_criterion))

	def __repr__(self):
		return "#" + str(self.def_index) + ": " + self.def_text

	def match(self, token):
		potential_groups = []
		for def_item in self.defs:
			tok_value = getattr(token,def_item.criterion)
			match_obj = def_item.match_func(def_item,tok_value)

			if match_obj is None:
				return False
			elif not match_obj:
				return False
			elif match_obj is True:
				pass
			elif match_obj is not None:
				if len(match_obj.groups()) > 0:
					potential_groups.append(match_obj.groups())
		self.groups = potential_groups
		return True


class Definition:

	def __init__(self, criterion, value, negative=False):
		# Handle conllu criterion aliases:
		if criterion == "form":
			criterion = "text"
		elif criterion == "upostag":
			criterion = "pos"
		elif criterion == "xpostag":
			criterion = "cpos"
		elif criterion == "feats":
			criterion = "morph"
		elif criterion == "deprel":
			criterion = "func"
		elif criterion == "deps":
			criterion = "head2"
		elif criterion == "misc":
			criterion = "func2"

		self.criterion = criterion
		self.value = value
		self.match_type = ""
		self.compiled_re = None
		self.negative = negative
		self.set_match_type()

	def set_match_type(self):
		value = self.value[1:-1]

		if self.value == "^.*$" and not self.negative:
			self.match_func = self.return_true
		elif re.escape(value) == value: # No regex operators within  expression
			if self.negative:
				self.match_func = self.return_exact_negative
			else:
				self.match_func = self.return_exact
			self.value = value
		else:  # regex
			self.compiled_re = re.compile(self.value)
			if self.negative:
				self.match_func = self.return_regex_negative
			else:
				self.match_func = self.return_regex

	@staticmethod
	def return_exact(definition,test_val):
		return test_val == definition.value

	@staticmethod
	def return_exact_negative(definition,test_val):
		return test_val != definition.value

	@staticmethod
	def return_regex(definition,test_val):
		return definition.compiled_re.search(test_val)

	@staticmethod
	def return_regex_negative(definition,test_val):
		return definition.compiled_re.search(test_val) is None

	@staticmethod
	def return_true(definition,test_val):
		return True


class Match:

	def __init__(self, def_index, token, groups):
		self.def_index = def_index
		self.token = token
		self.groups = groups

	def __repr__(self):
		return "#" + str(self.def_index) + ": " + self.token.__repr__


class DepEdit():

	def __init__(self, config_file="", options=None):
		self.transformations = []
		self.user_transformation_counter = 0
		self.quiet = False
		self.kill = None
		if options is not None:
			if "quiet" in options.__dict__ or options.quiet is not None:
				self.quiet = options.quiet
			if "kill" in options.__dict__ or options.kill is not None:
				if options.kill in ["supertoks","comments","both"]:
					self.kill = options.kill
		if not config_file == "":
			self.read_config_file(config_file)

	def read_config_file(self, config_file, clear_transformations = False):
		"""
		Function to read configuration file. Can be invoked after initialization.

		:param config_file: The file to read instructions from.
		:param clear_transformations: Whether to discard previous instructions.
		:return: void
		"""

		if clear_transformations:
			self.transformations = []
			self.user_transformation_counter = 0
		line_num = 0
		if isinstance(config_file, str):
			if sys.version_info[0] < 3:
				config_file = open(config_file).readlines()
			else:
				config_file = open(config_file,encoding="utf8").readlines()
		for instruction in config_file:
			line_num += 1
			if len(instruction)>0 and not instruction.startswith(";") and not instruction.startswith("#") and not instruction.strip() =="":
				self.transformations.append(Transformation(instruction, line_num))

		trans_report = ""
		for transformation in self.transformations:
			temp_report = transformation.validate()
			if temp_report != "":
				trans_report += "On line " + str(transformation.line) + ": " + temp_report +"\n"
		if len(trans_report) > 0:
			trans_report = "Depedit says: error in configuration file\n\n" + trans_report
			sys.stderr.write(trans_report)
			sys.exit()

	def process_sentence(self, conll_tokens, tokoffset, supertok_offset, transformations):
		for transformation in transformations:
			node_matches = defaultdict(list)
			for def_matcher in transformation.definitions:
				for token in conll_tokens[tokoffset+supertok_offset+1:]:
					if not token.is_super_tok:
						if def_matcher.match(token):
							node_matches[def_matcher.def_index].append(Match(def_matcher.def_index, token, def_matcher.groups))
			result_sets = []
			for relation in transformation.relations:
				found = self.matches_relation(node_matches, relation, result_sets)
				if not found:
					result_sets = []
			result_sets = self.merge_sets(result_sets,len(transformation.definitions),len(transformation.relations))
			self.add_groups(result_sets)
			if len(result_sets) > 0:
				for action in transformation.actions:
					retval = self.execute_action(result_sets, action)
					if retval == "last": # Explicit instruction to cease processing
						return self.serialize_output_tree(conll_tokens[tokoffset + supertok_offset + 1:], tokoffset)
		return self.serialize_output_tree(conll_tokens[tokoffset + supertok_offset + 1:], tokoffset)

	def matches_relation(self, node_matches, relation, result_sets):
		if len(relation) == 0:
			return
		elif "==" in relation:
			m = re.search(r':(.+)==', relation)
			operator = m.group()
			field = m.group(1)
		elif "." in relation:
			if re.match(r'.*\.[0-9]', relation):
				m = re.match(r'.*\.[0-9]*,?[0-9]*#', relation)
				operator = m.group()
				operator = operator[operator.find("."):operator.rfind("#")]
			else:
				operator = "."
		elif ">" in relation:
			operator = ">"

		matches = defaultdict(list)

		hits=0
		if relation == "none": # Unary operation on one node
			node1 = 1
			for matcher1 in node_matches[node1]:
				tok1 = matcher1.token
				hits += 1
				result = {}
				matches[node1].append(tok1)
				result[node1] = tok1
				result["rel"] = relation
				result["matchers"] = [matcher1]
				result_sets.append(result)
		elif "==" in relation:
			node1 = relation.split(operator)[0]
			node2 = relation.split(operator)[1]
			node1=int(node1.replace("#", ""))
			node2=int(node2.replace("#", ""))

			for matcher1 in node_matches[node1]:
				tok1 = matcher1.token
				for matcher2 in node_matches[node2]:
					tok2 = matcher2.token
					if self.test_relation(tok1, tok2, field):
						result_sets.append({node1: tok1, node2: tok2, "rel": relation, "matchers": [matcher1, matcher2] })
						matches[node1].append(tok1)
						matches[node2].append(tok2)
						hits += 1

			for option in [node1,node2]:
				matchers_to_remove = []
				for matcher in node_matches[option]:
					if matcher.token not in matches[option]:
						matchers_to_remove.append(matcher)
				for matcher in matchers_to_remove:
					node_matches[option].remove(matcher)
		else:
			node1 = relation.split(operator)[0]
			node2 = relation.split(operator)[1]

			node1=int(node1.replace("#", ""))
			node2=int(node2.replace("#", ""))
			for matcher1 in node_matches[node1]:
				tok1 = matcher1.token
				for matcher2 in node_matches[node2]:
					tok2 = matcher2.token
					if self.test_relation(tok1, tok2, operator):
						result_sets.append({node1: tok1, node2: tok2, "rel": relation, "matchers": [matcher1, matcher2] })
						matches[node1].append(tok1)
						matches[node2].append(tok2)
						hits += 1

			for option in [node1,node2]:
				matchers_to_remove = []
				for matcher in node_matches[option]:
					if matcher.token not in matches[option]:
						matchers_to_remove.append(matcher)
				for matcher in matchers_to_remove:
					node_matches[option].remove(matcher)

		if hits == 0:  # No solutions found for this relation
			return False
		else:
			return True

	@staticmethod
	def test_relation(node1,node2,operator):
		if operator == ".":
			if int(float(node2.id)) == int(float(node1.id))+1:
				return True
			else:
				return False
		elif operator == ">":
			if int(float(node2.head)) == int(float(node1.id)):
				return True
			else:
				return False
		elif "." in operator:
			m = re.match(r'\.([0-9]+)(,[0-9]+)?',operator)
			if len(m.groups()) > 1:
				min_dist = int(m.group(1))
				if not m.group(2) is None:
					max_dist = int(m.group(2).replace(",",""))
				else:
					max_dist = min_dist
				if max_dist >= int(float(node2.id)) - int(float(node1.id)) >= min_dist:
					return True
				else:
					return False
			else:
				dist = int(m.group(1))
				if int(float(node2.id)) - int(float(node1.id)) == dist:
					return True
				else:
					return False
		else:
			val1 = getattr(node1,operator)
			val2 = getattr(node2,operator)
			return val1 == val2


	def merge_sets(self, sets, node_count, rel_count):

		solutions = []
		bins = []
		for set_to_merge in sets:
			new_set = {}
			new_set["rels"] = []
			new_set["matchers"] = []
			for key in set_to_merge:
				if key == "rel":
					new_set["rels"].append(set_to_merge[key])
				elif key == "matchers":
					new_set["matchers"] += set_to_merge[key]
				else:
					new_set[key] = set_to_merge[key]

			for my_bin in copy(bins):
				if self.bins_compatible(new_set, my_bin):
					candidate = self.merge_bins(new_set, my_bin)
					bins.append(copy(candidate))
			bins.append(copy(new_set))

		for my_bin in bins:
			if len(my_bin) == node_count + 2:
				if len(my_bin["rels"]) == rel_count:  # All required relations have been fulfilled
					solutions.append(my_bin)
				else:  # Some node pair has multiple relations, check that all are fulfilled
					for set_to_merge in sets:
						if set_to_merge["rel"] not in my_bin["rels"]: # This relation was missing
							node_ids = list((key) for key in set_to_merge if isinstance(key,int))
							# Check that the missing rel connects nodes that are already in my_bin
							ids_are_in_bin = list((nid in my_bin) for nid in node_ids)
							if all(ids_are_in_bin):
								nodes_are_identical = list((set_to_merge[nid] == my_bin[nid]) for nid in node_ids)
								if all(nodes_are_identical):
									my_bin["rels"].append(set_to_merge["rel"])
									if len(my_bin["rels"]) == rel_count:  # Check if we now have all required relations
										solutions.append(my_bin)

		merged_bins = []
		for solution in solutions:
			self.merge_solutions(solution,merged_bins,rel_count)
		self.prune_merged_bins(merged_bins,rel_count)
		return merged_bins


	@staticmethod
	def merge_solutions(solution, merged, rel_count):
		merges_to_add = []
		if solution not in merged:
			merged.append(solution)
		if len(solution["rels"]) == rel_count:
			# This solution is completed
			pass
		else:
			# This is an incomplete solution, try to merge it into the merged solutions list
			for candidate in merged:
				if candidate != solution:
					for key in solution:
						if key != "rels" and key !="matchers":
							if key in candidate:  # This is a position, e.g. #1, that is also in the candidate in merged
								if solution[key] == candidate[key]:  # Check that they are compatible, e.g. #1 is the same token
									rels_in_candidate = (rel in candidate["rels"] for rel in solution["rels"])
									if not all(rels_in_candidate):
										rels = solution["rels"]+candidate["rels"]
										matchers = []
										for matcher in solution["matchers"]:
											matchers.append(matcher)
										for matcher in candidate["matchers"]:
											if matcher not in matchers:
												matchers.append(matcher)
										merged_solution = copy(solution)
										merged_solution.update(candidate)
										merged_solution["rels"] = rels
										merged_solution["matchers"] = matchers
										merges_to_add.append(merged_solution)
		if len(merges_to_add)>0:
			for merge_to_add in merges_to_add:
				merged.append(merge_to_add)
		solution["rels"].sort()


	@staticmethod
	def bins_compatible(bin1, bin2):
		overlap = False
		non_overlap = False
		for key in bin1:
			if key in bin2:
				if bin1[key] == bin2[key]:
					overlap = True
			if key not in bin2:
				non_overlap = True
		if overlap and non_overlap:
			return True
		else:
			return False

	@staticmethod
	def merge_bins(bin1, bin2):
		"""
		Merge bins we know are compatible, e.g. bin1 has #1+#2 and bin2 has #2+#3

		:param bin1: a bin dictionary mapping indices to tokens, a list of relations 'rels' and matcher objects 'matchers'
		:param bin2: a bin dictionary mapping indices to tokens, a list of relations 'rels' and matcher objects 'matchers'
		:return: the merged bin with data from both input bins
		"""
		for matcher in bin1["matchers"]:
			skip = False
			for matcher2 in bin2["matchers"]:
				if matcher2.def_index == matcher.def_index:
					skip = True
			if not skip:
				bin2["matchers"].append(matcher)
		for key in bin1:
			if key != "rels":
				if key not in bin2:
					out_bin = copy(bin2)
					out_bin[key]= bin1[key]
					for rel in bin1["rels"]:
						new_rel = deepcopy(rel)
						out_bin["rels"] = bin2["rels"]+[new_rel]
					return out_bin


	@staticmethod
	def prune_merged_bins(merged_bins,rel_count):
		"""
		Deletes bins with too few relationships matched after merging is complete

		:param merged_bins: candidates for bins representing complete related chains of nodes
		:param rel_count: how many relations the current transformation has - any bins with less will be discarded now
		:return: void
		"""
		bins_to_delete = []
		for merged_bin in merged_bins:
			if len(merged_bin["rels"]) < rel_count:
				bins_to_delete.append(merged_bin)
		for bin_to_delete in bins_to_delete:
			merged_bins.remove(bin_to_delete)

	@staticmethod
	def add_groups(result_sets):
		for result in result_sets:
			groups = []
			sorted_matchers = sorted(result["matchers"], key=lambda x: x.def_index)
			for matcher in sorted_matchers:
				if len(matcher.groups) > 0:
					for group in matcher.groups:
						groups.append(group[0])
			result["groups"] = groups[:]

	@staticmethod
	def execute_action(result_sets, action_list):
		actions = action_list.split(";")
		for result in result_sets:
			if len(result) > 0:
				for action in actions:
					if action == "last":
						return "last"
					elif ":" in action:  # Unary instruction
						if action.startswith("#S:"):  # Sentence annotation instruction
							key_val = action.split(":")[1]
							key, val = key_val.split("=",1)
							result[1].sentence.annotations[key] = val
						else: # node instruction
							node_position = int(action[1:action.find(":")])
							property = action[action.find(":")+1:action.find("=")]
							value = action[action.find("=")+1:].strip()
							group_num_matches = re.findall(r"(\$[0-9]+[LU]?)",value)
							if group_num_matches is not None:
								for g in group_num_matches:
									no_dollar = g[1:]
									case = ""
									if no_dollar[-1] == "U":
										case = "upper"
										no_dollar = no_dollar[0:-1]
									elif no_dollar[-1] == "L":
										case = "lower"
										no_dollar = no_dollar[0:-1]
									group_num = int(no_dollar)
									try:
										group_value = result["groups"][group_num - 1]
										if case == "lower":
											group_value = group_value.lower()
										elif case == "upper":
											group_value = group_value.upper()
									except IndexError:
										sys.stderr.write("The action '" + action + "' refers to a missing regex bracket group '$" + str(group_num) + "'\n")
										sys.exit()
									group_str = str(group_num)
									if case == "lower":
										group_str += "L"
									elif case == "upper":
										group_str += "U"
									value = re.sub(r"\$"+group_str,group_value,value)
							setattr(result[node_position],property,value)
					else:  # Binary instruction
						if ">" in action:  # Head relation
							operator = ">"
							node1 = int(action.split(operator)[0].replace("#", ""))
							node2 = int(action.split(operator)[1].replace("#", ""))
							tok1 = result[node1]
							tok2 = result[node2]
							if tok1 != tok2:
								tok2.head = tok1.id

	def add_transformation(self, *args, **kwargs):
		"""
		Flexible function for adding transformations to an imported DepEdit object, rather than reading them from a configuration file

		:param args: a string specifying a transformation line (three specs separated by two tabs), or a tuples of such strings
		:param kwargs: alternatively, a dictionary supplying three lists for the keys: 'nodes', 'rels', 'actions'
		:return: void
		"""

		if len(args)>0:  # Single string transformation(s)

			if not isinstance(args,list): # Make list out of input tuple
				args = list(arg for arg in args)
				if isinstance(args[0], list):  # Flatten list in case user provided a list
					args = args[0]
			for transformation_string in args:
				try:
					self.user_transformation_counter += 1
					user_line_number = "u" + str(self.user_transformation_counter)
					new_transformation = Transformation(transformation_string,user_line_number)
					self.transformations.append(new_transformation)
				except:
					raise IOError("Invalid transformation - must be a string")
		elif "nodes" in kwargs and "rels" in kwargs and "actions" in kwargs:
			if not (isinstance(kwargs["nodes"], list) and (isinstance(kwargs["nodes"], list)) and (isinstance(kwargs["nodes"], list))):
				raise IOError("Invalid transformation - must supply three lists in values for keys: nodes, rels, actions")
			else:
				node_clause = ";".join(kwargs["nodes"])
				rel_clause = ";".join(kwargs["rels"])
				actions_clause = ";".join(kwargs["actions"])
				transformation_string = "\t".join([node_clause,rel_clause,actions_clause])

				self.user_transformation_counter += 1
				user_line_number = "u" + str(self.user_transformation_counter)
				new_transformation = Transformation(transformation_string, user_line_number)
				self.transformations.append(new_transformation)

	def serialize_output_tree(self,tokens, tokoffset):
		output_tree = ""
		for tok in tokens:
			if tok.is_super_tok:
				if self.kill in ["both","supertoks"]:
					continue
				tok_head_string = tok.head
				tok_id = tok.id
			elif tok.head == "0":
				tok_head_string = "0"
				tok_id = str(float(tok.id) - tokoffset)
			else:
				tok_head_string = str(float(tok.head)-tokoffset)
				tok_id = str(float(tok.id)-tokoffset)
			# Only keep decimal ID component for non-0 ellipsis IDs, e.g. 10.1 - those tokens have normal head '_'
			tok_id = tok_id.replace(".0","")
			tok_head_string = tok_head_string.replace(".0","")
			if "." in tok_id:
				tok_head_string = "_"
			if self.input_mode == "8col":
				output_tree += tok_id+"\t"+tok.text+"\t"+tok.lemma+"\t"+tok.pos+"\t"+tok.cpos+"\t"+tok.morph+\
								"\t"+tok_head_string+"\t"+tok.func+"\n"
			else:
				output_tree += tok_id+"\t"+tok.text+"\t"+tok.lemma+"\t"+tok.pos+"\t"+tok.cpos+"\t"+tok.morph+\
								"\t"+tok_head_string+"\t"+tok.func+"\t"+tok.head2+"\t"+tok.func2+"\n"
		return output_tree

	def make_sent_id(self, sent_id):
		return "# sent_id = " + self.docname + "-" + str(sent_id) + "\n"

	def run_depedit(self, infile, filename="file", sent_id=False, docname=False):
		children = defaultdict(list)
		child_funcs = defaultdict(list)
		conll_tokens = []
		self.input_mode = "10col"
		self.docname = filename
		tokoffset = 0
		sent_num = 0
		supertok_offset = 0
		sentlength = 0
		supertok_length = 0

		conll_tokens.append(0)
		my_output = ""
		sentence_string = ""
		current_sentence = Sentence()

		# Check if DepEdit has been fed an unsplit string programmatically
		if isinstance(infile,str):
			infile = infile.split("\n")

		for myline in infile:
			if sentlength > 0 and "\t" not in myline:
				current_sentence.length = sentlength
				conll_tokens[-1].position = "last"
				transformed = self.process_sentence(conll_tokens, tokoffset, supertok_offset, self.transformations)
				transformed = current_sentence.print_annos() + transformed
				sent_num += 1
				if sent_id:
					my_output += self.make_sent_id(sent_num)
				my_output += transformed
				sentence_string = ""
				current_sentence = Sentence()
				if sentlength > 0:
					tokoffset += sentlength
					supertok_offset += supertok_length
				sentlength = 0
				supertok_length = 0
			if myline.startswith("#"):  # Preserve comment lines unless kill requested
				if self.kill not in ["comments","both"]:
					my_output += myline.strip() + "\n"
			elif myline.strip() == "":
				my_output += "\n"
			elif myline.find("\t") > 0:  # Only process lines that contain tabs (i.e. conll tokens)
				sentence_string += myline
				cols = myline.split("\t")
				if "-" in cols[0]: # potential conllu super-token, just preserve
					super_tok = True
					tok_id = cols[0]
					head_id = cols[6]
				else:
					super_tok = False
					tok_id = str(float(cols[0]) + tokoffset)
					if cols[6] == "_":
						if not self.quiet:
							sys.stderr.write("DepEdit WARN: head not set for token " + tok_id + " in " + filename + "\n")
						head_id = str(0 + tokoffset)
					else:
						head_id = str(float(cols[6]) + tokoffset)
				if len(cols) > 8:
					# Collect token from line; note that head2 is parsed as a string, which is often "_" for monoplanar trees
					this_tok = ParsedToken(tok_id, cols[1], cols[2], cols[3], cols[4], cols[5],head_id, cols[7].strip(), cols[8], cols[9].strip(), cols[0], [], "mid",super_tok)
				else:  # Attempt to read as 8 column Malt input
					this_tok = ParsedToken(tok_id,cols[1],cols[2],cols[3],cols[4],cols[5],head_id,cols[7].strip(),cols[6],cols[7].strip(),cols[0],[], "mid",super_tok)
					self.input_mode = "8col"
				if cols[0] == "1" and not super_tok:
					this_tok.position = "first"
				this_tok.sentence = current_sentence
				conll_tokens.append(this_tok)
				if not super_tok:
					sentlength += 1
					children[str(float(head_id) + tokoffset)].append(tok_id)
					child_funcs[(float(head_id) + tokoffset)].append(cols[7])
				else:
					supertok_length += 1

		if sentlength > 0:  # Possible final sentence without trailing new line
			current_sentence.length = sentlength
			conll_tokens[-1].position = "last"
			transformed = self.process_sentence(conll_tokens, tokoffset, supertok_offset, self.transformations)
			transformed = current_sentence.print_annos() + transformed
			sent_num += 1
			if sent_id:
				my_output += self.make_sent_id(sent_num)
			my_output += transformed

		if docname:
			newdoc = '# newdoc id = ' + self.docname + "\n"
			my_output = newdoc + my_output

		return my_output

if __name__ == "__main__":
	depedit_version = "DepEdit V" + __version__
	parser = argparse.ArgumentParser()
	parser.add_argument('file',action="store",help="Input single file name or glob pattern to process a batch (e.g. *.conll10)")
	parser.add_argument('-c', '--config', action="store", dest="config", default="config.ini", help="Configuration file defining transformation")
	parser.add_argument('-d', '--docname', action="store_true", dest="docname", help="Begin output with # newdoc id =...")
	parser.add_argument('-s', '--sent_id', action="store_true", dest="sent_id", help="Add running sentence ID comments")
	parser.add_argument('-k', '--kill', action="store", choices=["supertoks","comments","both"], help="Remove supertokens or commments from output")
	parser.add_argument('-q', '--quiet', action="store_true", dest="quiet", help="Do not output warnings and messages")
	group = parser.add_argument_group('Batch mode options')
	group.add_argument('-o', '--outdir', action="store", dest="outdir", default="", help="Output directory in batch mode")
	group.add_argument('-e', '--extension', action="store", dest="extension", default="", help="Extension for output files in batch mode")
	group.add_argument('-i', '--infix', action="store", dest="infix", default=".depedit", help="Infix to denote edited files in batch mode (default: .depedit)")
	parser.add_argument('--version', action='version', version=depedit_version)
	options = parser.parse_args()

	if options.extension.startswith("."):  # Ensure user specified extension does not include leading '.'
		options.extension = options.extension[1:]

	try:
		config_file = io_open(options.config, encoding="utf8")
	except IOError:
		sys.stderr.write("\nConfiguration file not found (specify with -c or use the default 'config.ini')\n")
		sys.exit()
	depedit = DepEdit(config_file=config_file,options=options)

	if sys.platform == "win32":  # Print \n new lines in Windows
		import os, msvcrt
		msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

	files = glob(options.file)
	for filename in files:
		infile = io_open(filename, encoding="utf8")
		basename = os.path.basename(filename)
		if options.docname or options.sent_id:
			docname = basename[:basename.rfind(".")]
		else:
			docname = filename
		output_trees = depedit.run_depedit(infile,docname, sent_id=options.sent_id, docname=options.docname)
		if len(files) == 1:
			# Single file being processed, just print to STDOUT
			if sys.version_info[0] < 3:
				print(output_trees.encode("utf-8"),end="")
			else:
				print(output_trees,end="")
		else:
			# Multiple files, add '.depedit' or other infix from options before extension and write to file
			if options.outdir != "":
				if not options.outdir.endswith(os.sep):
					options.outdir += os.sep
			outname = options.outdir + basename
			if "." in filename:
				extension = outname[outname.rfind(".")+1:]
				if options.extension != "":
					extension = options.extension
				outname = outname[:outname.rfind(".")]
				outname += options.infix + "." + extension
			else:
				if options.extension == "":
					outname += options.infix
				else:
					outname += options.infix + "." + options.extension
			if sys.version_info[0] < 3:
				with open(outname,'wb') as f:
					f.write(output_trees.encode("utf-8"))
			else:
				with open(outname,'w',encoding="utf8") as f:
					f.write(output_trees.encode("utf-8"))




import sys, io, re, os
from argparse import ArgumentParser
from f_score_segs import main as f_score
from collections import defaultdict
from six import iterkeys

PY3 = sys.version_info[0] == 3

script_dir = os.path.dirname(os.path.realpath(__file__)) + os.sep

lex = script_dir + ".." + os.sep + "data" + os.sep + "copt_lemma_lex_cplx_2.5.tab"
frq = script_dir + ".." + os.sep + "data" + os.sep + "cop_freqs.tab"
conf = script_dir + ".." + os.sep + "data" + os.sep + "test.conf"
ambig = script_dir + ".." + os.sep + "data" + os.sep + "ambig.tab"

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'lib')))

from tokenize_rf import RFTokenizer
from stacked_tokenizer import StackedTokenizer


def tt2seg_table(tt_string,group_attr="norm_group",unit_attr="norm"):

	group = ""
	output = ""
	units = []
	pairs = []
	for line in tt_string.split("\n"):
		if " " + group_attr + "=" in line:  # new group
			group = re.search(" " + group_attr + '="([^"]+)"',line).group(1)
		if "</" + group_attr +">" in line:
			pairs.append((group,"|".join(units)))
			group = ""
			units = []
		if " " + unit_attr + "=" in line:  # new unit
			unit = re.search(" " + unit_attr + '="([^"]+)"',line).group(1)
			units.append(unit)

	for grp, segs in pairs:
		output += grp + "\t" + segs + "\n"

	return output


p = ArgumentParser()
p.add_argument("--train_list",default="train_list.tab",help="file with one file name per line of TT SGML training files")
p.add_argument("--test_list",default="test_list.tab",help="file with one file name per line of TT SGML test files")

opts = p.parse_args()

train_list = io.open(opts.train_list,encoding="utf8").read().strip().split("\n")
test_list = io.open(opts.test_list,encoding="utf8").read().strip().split("\n")

train = ""
for file_ in train_list:
	tt_sgml = io.open(script_dir + file_,encoding="utf8").read()
	train += tt2seg_table(tt_sgml)

test = ""
for file_ in test_list:
	tt_sgml = io.open(script_dir + file_,encoding="utf8").read()
	test += tt2seg_table(tt_sgml)

# Remove bug rows
clean_test = []
for line in test.strip().split("\n"):
	grp, seg = line.split("\t")
	if len(grp) == len(seg.replace("|","")):
		clean_test.append(line)
test = "\n".join(clean_test)


rf = RFTokenizer(model="test")
if not PY3:
	train = unicode(train)


#io_file = io.StringIO()
#io_file.write(train)
#io_file.flush()
#io_file.close()

with io.open("_tmp_train.tab",'w', encoding="utf8") as f:
	f.write(train)

with io.open("_tmp_test.tab",'w', encoding="utf8") as f:
	f.write(test)

test_input = ""
for line in test.strip().split("\n"):
	test_input += line.split("\t")[0] + "\n"

golds = test.split("\n")

retrain=False
if retrain:
	rf.train("_tmp_train.tab",lexicon_file=lex,freq_file=frq,test_prop=0,dump_model=True,output_errors=True,conf=conf)
	preds = rf.rf_tokenize(test_input.strip().split("\n"))
	f_score("_tmp_test.tab","\n".join(preds),preds_as_string=True,ignore_diff_len=True)


stk = StackedTokenizer(no_morphs=True,model="test",pipes=True)
stk.load_ambig(ambig_table=ambig)

preds = stk.analyze("_".join(test_input.strip().split("\n")))
preds = preds.split("_")


f_score("_tmp_test.tab","\n".join(preds),preds_as_string=True,ignore_diff_len=True)

errs = defaultdict(int)
for i, pred in enumerate(preds):
	gold = golds[i].split("\t")[1]
	if pred != gold:
		errs[gold + "\t"+ pred] += 1

with io.open("errs.tab",'w',encoding="utf8") as f:
	for key in sorted(iterkeys(errs),reverse=True):
		f.write(key + "\t" + str(errs[key]) + "\n")


from trankit import TPipeline

trainer = TPipeline(training_config={
    'task': 'posdep',
    'category':'customized',
    'save_dir': './trankit/saved_model',
    #'train_txt_fpath': './trankit/train.txt',
    'train_conllu_fpath': './trankit/he_htb-ud-train.conllu',
    #'dev_txt_fpath': './trankit/dev.txt',
    'dev_conllu_fpath': './trankit/he_htb-ud-dev.conllu'
    }
)

trainer.train()


#!/usr/bin/python
# -*- coding: utf-8 -*-

########################################################################
#                                                                      #
#  tokenization script for tagger preprocessing                        #
#  Adapted from TreeTagger tokenizer in Perl by:                       #
#          Helmut Schmid, IMS, University of Stuttgart                 #
#          Serge Sharoff, University of Leeds                          #
#  Description:                                                        #
#  - splits input text into tokens (one token per line)                #
#  - cuts off punctuation, parentheses etc.                            #
#  - disambiguates periods                                             #
#  - preserves SGML markup                                             #
#                                                                      #
#  - Ported to Python and adapted to Hebrew by Amir Zeldes             #
########################################################################


from argparse import ArgumentParser
from depedit import DepEdit
import io, sys, re

PY3 = sys.version_info[0] > 2

# characters which have to be cut off at the beginning of a word
PChar = r"""\[¿¡{(`"‚„†‡‹‘’“”•'–—›«「"""

# characters which have to be cut off at the end of a word
FChar = r"""'\]}'`"),;:!?%‚„…†‡‰‹‘’“”•–—›'»」"""

# character sequences which have to be cut off at the beginning of a word
PClitic = ""

# character sequences which have to be cut off at the end of a word
FClitic = ""


def tokenize(text, abbr=None, add_sents=False, from_pipes=False, aggressive_hyphenation=True):

    if aggressive_hyphenation:
        text = re.sub(r'(?<=[א-ת])([–/־-])(?=[א-ת])',r' \1 ', text)
        text = re.sub(r'(?<= [0-9])([–:־-])(?=[0-9] )',r' \1 ', text)  # sports scores etc.
        text = re.sub(r'(?<= (?:19|20)[0-9]{2})([–־-])(?=(?:19|20)[0-9]{2}[ ,\.])',r' \1 ', text)  # year ranges
        text = re.sub(r'(?<= [0-9][0-9])([–־-])(?=[0-9]+%)',r' \1 ', text)  # percentage ranges
        text = re.sub(r'( (?:(?:\d{1,3},)+\d{3}))([–־-])(?=(?:(?:\d{1,3},)+\d{3})[ ,\.])',r'\1 \2 ', text)  # ranges of numbers with thousands sep

    output = ""
    if add_sents:
        lines = []
        raw_lines = text.split("\n")
        for line in raw_lines:
            if len(line.strip())>0:
                lines.append("<s>" + line + "</s>")
    else:
        lines = text.split("\n")

    if from_pipes:  # Text is already whitespace tokenized
        data = "\n".join(lines)
        data = data.replace("<s>","\n<s>\n").replace("</s>","\n</s>\n")
        data = re.sub(r'\n+',r'\n',data)
        return "\n".join(data.split())

    # Read the list of abbreviations and words
    Token = set([])
    if abbr is not None:
        abbr_lines = io.open(abbr,encoding="utf8").read().replace("\r","").strip().split("\n")
        for line in abbr_lines:
            Token.add(line)

    for line in lines:
        # replace newlines and tab characters with blanks
        line = line.replace("\t"," ").replace("\n"," ")

        # replace blanks within SGML tags
        sep1 = "□"
        sep2 = "■"

        if not PY3:
            sep1 = sep1.decode("utf8")
            sep2 = sep2.decode("utf8")

        find_tag_space = r'(<[^<> ]*) ([^<>]*>)'
        while re.search(find_tag_space,line) is not None:
            line = re.sub(find_tag_space,r'\1'+sep1+r'\2',line)

        # replace whitespace with a special character
        line = line.replace(" ",sep2)

        # restore SGML tags
        line = line.replace(sep1," ")
        line = line.replace(sep2,sep1)

        # prepare SGML-Tags for tokenization

        line = re.sub(r'(<[^<>]*>)',sep1 + r"\1" + sep1,line)
        line = re.sub(r'^' + sep1,"",line)
        line = re.sub(sep1 + r'$',"",line)
        line = re.sub(sep1*3 +"*",sep1,line)

        units = line.split(sep1)
        for i, unit in enumerate(units):
            if re.match(r"<.*>$",unit) is not None:
                # SGML tag
                output += unit + "\n"
            else:
                #add a blank at the beginning and the end of each segment
                email = r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
                url = r'https?://(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&/=]*)'
                acro = r"([A-Z]\.([A-Z]\.)+)"
                short_url = r"[-a-zA-Z0-9]{2,256}\.(com|org|edu|co|io|net)"

                # insert missing blanks after punctuation if not an abbreviation
                if unit not in Token and re.search(email,unit) is None and re.search(url,unit) is None \
                        and re.search(acro, unit) is None and re.search(short_url, unit) is None:
                    unit = " " + unit + " "
                    unit = re.sub(r'\.\.\.'," ... ",unit)
                    unit = re.sub(r'([;!?])([^ ])',r'\1 \2', unit)
                    unit = re.sub(r'([.,:])([^ 0-9.])', r'\1 \2',unit)
                else:
                    unit = " " + unit + " "

                subunits = unit.split()
                for subunit in subunits:
                    suffix=""

                    # separate punctuation and parentheses from words
                    while True:
                        finished = True
                        # cut off preceding punctuation
                        m = re.match(r'(['+PChar+'])(.)',subunit)
                        if m is not None:
                            m1 = m.group(1)
                            subunit = re.sub('^(['+PChar+'])(.)',r'\2',subunit)
                            output += m1 + "\n"
                            finished = 0
                        # cut off trailing punctuation
                        m = re.search(r'(.)(['+FChar+'])$',subunit)
                        if m is not None:
                            m2 = m.group(2)
                            subunit = re.sub(r'(.)(['+FChar+'])$',r'\1',subunit)
                            suffix = m2 + "\n" + suffix
                            finished = 0

                        # cut off trailing periods if punctuation precedes
                        m = re.search(r'(['+FChar+'])\.$',subunit)
                        if m is not None:
                            subunit = re.sub(r'(['+FChar+'])\.$','',subunit,count=1)
                            suffix = ".\n" + suffix
                            if subunit == "":
                                subunit = m.group(1)
                            else:
                                suffix = m.group(1) + "\n" + suffix
                            finished = False
                        if finished:
                            break

                    # handle explicitly listed tokens
                    if subunit in Token:
                        output += subunit + "\n" + suffix
                        continue

                    # abbreviations of the form A. or U.S.A.
                    if re.match(r'([A-Za-z-]\.)+$',subunit) is not None:
                        output += subunit + "\n" + suffix
                        continue

                    # e-mail addresses
                    if re.search(email,subunit) is not None or re.search(url,subunit) is not None:
                        output += subunit + "\n" + suffix
                        continue

                    # disambiguate periods
                    m = re.match(r'(..*)\.$',subunit)
                    if m is not None and subunit != "...":
                        subunit = m.group(1)
                        suffix = ".\n" + suffix
                        if subunit in Token:
                            output += subunit + "\n" + suffix
                            continue

                    # cut off clitics
                    while re.match(r'(--)(.)',subunit) is not None:
                        m = re.match(r'(--)(.)',subunit)
                        subunit = re.sub(r'(--)(.)',r'\2',subunit)
                        output += m.group(1) + "\n"

                    if PClitic != '':
                        while re.match(r'('+PClitic+')(.)',subunit) is not None:
                            m = re.match(r'('+PClitic+')(.)',subunit)
                            subunit = re.sub(r'('+PClitic+')(.)',r'\2',subunit)
                            output += m.group(1) + "\n"

                    while re.search(r'(.)(--)$',subunit) is not None:
                        m = re.search(r'(.)(--)$',subunit)
                        subunit = re.sub(r'(.)(--)$',r'\1',subunit)
                        suffix = m.group(2) + "\n" + suffix

                    if FClitic != "":
                        while re.search(r'(.)('+FClitic+')$',subunit) is not None:
                            m = re.search(r'(.)('+FClitic+')$',subunit)
                            subunit = re.sub(r'(.)('+FClitic+')$',r"\1",subunit)
                            suffix = m.group(2) + "\n" + suffix
                    output+=subunit + "\n" + suffix
    return output


def add_feat(morph, feat):
    if morph == "_":
        return feat
    else:
        attrs = morph.split("|")
        attrs.append(feat)
        return "|".join(sorted(list(set(attrs))))


def move_binyan(conllu):
    # Move Binyan to Misc
    binyaned = []
    for line in conllu.split("\n"):
        if "HebBinyan" in line:
            fields = line.split("\t")
            binyan = re.search(r'(HebBinyan=[A-Z]+)', line).group(1)
            fields[5] = re.sub(r'HebBinyan=([A-Z]+)\|?', '', fields[5])
            fields[-1] = add_feat(fields[-1], binyan)
            line = "\t".join(fields)
        binyaned.append(line)
    return "\n".join(binyaned)


def add_space_after(plain_text, conllu):
    """Adds the SpaceAfter=No feature to conllu data based on original plain text, ignoring XML"""
    def chomp(plain_text, token):
        space_prop = ""
        token_index = plain_text.find(token)
        plain_text = plain_text[token_index+len(token):]
        in_xml = False
        for c in plain_text:
            if c in [" ","\n","\t","\r"]:
                space_prop = ""
                break
            elif c == "<":
                in_xml = True
            elif c == ">":
                in_xml = False
            elif not in_xml:
                # Non-whitespace character outside of XML, there was no space after the token
                space_prop = "SpaceAfter=No"
                break
        return space_prop, plain_text

    plain_text = re.sub(r'<[^<>]+>',"",plain_text)  # Destroy XML
    output = []
    skip = 0
    lines = conllu.split("\n")
    for line in lines:
        if "\t" in line:
            fields = line.split("\t")
            if "-" in fields[0]:  # Supertoken
                start, end = fields[0].split("-")
                skip = int(end) - int(start) + 1
                space_prop, plain_text = chomp(plain_text,fields[1])
                if space_prop != "":
                    fields[-1] = add_feat(fields[-1],space_prop)
                    line = "\t".join(fields)
            else:
                if skip > 0:
                    # This is inside a supertoken, no need to add SpaceAfter, just down-tick skip
                    skip -= 1
                else:
                    space_prop, plain_text = chomp(plain_text, fields[1])
                    if space_prop != "":
                        fields[-1] = add_feat(fields[-1],space_prop)
                        line = "\t".join(fields)
        output.append(line)

    output = "\n".join(output)
    #output = move_binyan(output)

    d = DepEdit(config_file=[])
    output = d.run_depedit(output,sent_text=True)

    return output


if __name__ == "__main__":

    p = ArgumentParser()
    p.add_argument("infile")
    p.add_argument("-a","--abbreviations",action="store",default=None,help="File name for list of abbreviations and other tokens to leave alone")

    opts = p.parse_args()

    input_text = io.open(opts.infile,encoding="utf8").read().replace("\r","").strip()
    output = tokenize(input_text,opts.abbreviations)
    print(output)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import io, sys, argparse

"""
Script to evaluate segmentation f-score and perfect super-token segmentation proportion from two files:
 * goldfile: single column file, one super-token perline, pipes denote segmentation positions	
	* predfile: same format
	
Stripped files are assumed to have the same amount of characters except pipes. A super-token may also consist
of a single pipe (i.e. input contained a literal pipe, which must be a complete super-token).
"""


def main(goldfile, predfile, preds_as_string=False,ignore_diff_len=False):
	lines = io.open(goldfile, encoding="utf8").read().strip().replace("\r", "").split("\n")
	counter = 0
	gold = []
	gold_groups = []
	perfect = 0
	total = 0

	if "\t" in lines[0]:  # Convenience step allowing 2-column file as gold
		sys.stderr.write("o Found tab in gold file, using second column as gold\n")
		lines = [line.split("\t")[1] for line in lines if "\t" in line]
	for r, line in enumerate(lines):
		total += 1
		gold_groups.append(line.strip())
		for i, c in enumerate(list(line.strip())):
			counter += 1
			if i == len(line.strip()) - 1:  # Last character is trivial, ignore
				continue
			if c == "|":
				if len(line.strip()) > 1 and i == 0:
					print("Complex token begins with boundary marker at gold row " + str(r))
					sys.exit()
				counter -=1
				gold[-1] = 1
			else:
				gold.append(0)

	gold_chars = counter

	if preds_as_string:
		lines = predfile.split("\n")
	else:
		lines = io.open(predfile, encoding="utf8").read().strip().replace("\r", "").split("\n")
	counter = 0
	pred = []

	to_pop = []
	for r, line in enumerate(lines):
		if len(line.replace("|","").strip()) != len(gold_groups[r].replace("|","").strip()):
			# Length mismatch, bug row
			if ignore_diff_len:
				continue
		if line.strip() == gold_groups[r]:
			perfect += 1
		for i, c in enumerate(list(line.strip())):
			counter += 1
			if i == len(line.strip()) - 1:  # Last character is trivial, ignore
				continue
			if c == "|":
				if len(line.strip()) > 1 and i == 0:
					print("Complex token begins with boundary marker at pred row " + str(r))
					sys.exit()
				counter -=1
				pred[-1] = 1
			else:
				pred.append(0)

	pred_chars = counter

	if pred_chars != gold_chars:
		sys.stderr.write("ERROR: found " + str(gold_chars) + " gold chars but " + str(pred_chars) + " pred chars\n")
		sys.exit()

	true_positive = 0
	false_positive = 0
	false_negative = 0
	for i in range(len(gold)):
		if gold[i] == 0:
			if pred[i] == 0:
				pass
			else:
				false_positive += 1
		else:
			if pred[i] == 0:
				false_negative += 1
			else:
				true_positive += 1

	try:
		precision = true_positive / (float(true_positive) + false_positive)
	except Exception as e:
		precision = 0

	try:
		recall = true_positive / (float(true_positive) + false_negative)
	except Exception as e:
		recall = 0

	try:
		f_score = 2 * (precision * recall) / (precision + recall)
	except:
		f_score = 0

	try:
		perf_score = perfect/float(total)
	except:
		perf_score = 0

	print("Total chars: " + str(pred_chars))
	print("Perfect groups: " + str(perf_score))
	print("Precision: " + str(precision))
	print("Recall: " + str(recall))
	print("F-Score: " + str(f_score))


if __name__ == "__main__":
	p = argparse.ArgumentParser()

	p.add_argument("goldfile")
	p.add_argument("predfile")

	opts = p.parse_args()

	main(opts.goldfile,opts.predfile)



import io, re
from collections import defaultdict

PRIORITIES = ["text","sp","table","row","cell","head","p","figure","caption","list","item","s","quote","q","hi","sic","ref",
			  "date","incident","w"]

OPEN_SGML_ELT = re.compile(r'^<([^/ ]+)( .*)?>$')
CLOSE_SGML_ELT = re.compile(r'^</([^/]+)>$')

class Span:

	def __init__(self,start=0,end=0,text="", elem="", priorities=PRIORITIES):
		self.start = start
		self.end = end
		self.text = text
		self.elem = elem
		self.length = end - start
		self.priority = priorities.index(elem) if elem in priorities else 100

	def __repr__(self):
		return str(self.start) + "-" + str(self.end) + ": " + self.text


def reorder(tt_sgml,priorities=PRIORITIES):
	# Preprocessing: if an element opens and closes immediately, the rest of the code will handle it incorrectly
	# unless we add a dummy token. It will be removed later.
	lines = tt_sgml.split("\n")
	i = 0
	while i < len(lines) - 1:
		line = lines[i]
		next_line = lines[i+1]
		open_match = re.search(OPEN_SGML_ELT, line)
		close_match = re.search(CLOSE_SGML_ELT, next_line)
		if open_match and close_match and open_match.groups()[0] == close_match.groups()[0]:
			lines.insert(i+1, "__SECRET_DUMMY_TOKEN__")
			i += 1
		i += 1

	# Pass 1: build data model
	open_elems = defaultdict(list)
	spans = []
	toknum = 1
	for line in lines:
		if line.startswith("</") and line.endswith(">"):  # Close element
			elem = re.search(r'^</([^\s>]*)',line).group(1)
			if elem not in open_elems:
				raise IOError("! saw a closed element: " + line + " but no corresponding element is open!\n")
			span = Span(start=open_elems[elem][-1][0],end=toknum,text=open_elems[elem][-1][1],elem=elem,priorities=priorities)
			open_elems[elem].pop()
			if len(open_elems[elem]) == 0:
				del open_elems[elem]
			spans.append(span)
		elif (line.startswith("<") and line.endswith("/>")) or line.startswith("<?"):  # Unary element, treat like token
			toknum += 1
		elif line.startswith("<") and line.endswith(">"):  # Open element
			elem = re.search(r'^<([^\s>]*)',line).group(1)
			open_elems[elem].append((toknum, line))
		elif len(line.strip()) > 0:
			toknum += 1

	# Build start/end dictionaries
	start_dict = defaultdict(list)
	end_dict = defaultdict(list)
	for span in spans:
		start_dict[span.start].append(span)
		end_dict[span.end].append(span)

	# Pass 2: reorder
	output = []
	toknum = 1
	for line in lines:
		if ((line.startswith("<") and line.endswith(">")) or len(line.strip()) == 0) and \
				not (line.startswith("<?") or (line.startswith("<") and line.endswith("/>"))):
			continue

		starting = start_dict[toknum]
		elems = sorted(starting, key=lambda x: (-x.length, x.priority))
		for elem in elems:
			output.append(elem.text)

		output.append(line)
		toknum += 1

		ending = end_dict[toknum]
		elems = sorted(ending, key=lambda x: (x.length, -x.priority))
		for elem in elems:
			output.append("</" + elem.elem + ">")

	return "\n".join(output).replace("__SECRET_DUMMY_TOKEN__\n", "") + "\n"


if __name__ == "__main__":
	test_sgml = """<?xml version="1.0" ?>
<text dateCollected="2019-11-05" dateCreated="2019-02-04" dateModified="2019-04-23" id="autogum_bio_doc031" shortTile="ingo-ruczinski" sourceURL="https://en.wikipedia.org/wiki/Ingo_Ruczinski" speakerCount="0" speakerList="none" title="Ingo Ruczinski" type="bio">
<head>
<s>
Awards
and
honors
</s>
</head>
<p>
<s>
In
2016
</s>
<s>
Ingo
Ruczinski
had
became
an
elected
fellow
of
the
<ref target="https://en.wikipedia.org/wiki/American_Statistical_Association">
American
Statistical
Association
</ref>
.
<ref target="https://en.wikipedia.org/wiki/Main_Page#cite_note-2">
[
2
]
</s>
</ref>
</p>
</text>"""

	print(reorder(test_sgml))

